from __future__ import annotations
import re

from ..util.seq import (MATCH, DELET, INS_5, INS_3,
                        SUB_A, SUB_C, SUB_G, SUB_T, SUB_N,
                        A_INT, C_INT, G_INT, T_INT, ANY_N)


class VectorError(Exception):
    """ Any error that occurs during vectoring """


class VectorValueError(VectorError, ValueError):
    """ Any ValueError that occurs during vectoring """


class VectorNotImplementedError(VectorError, NotImplementedError):
    """ Any NotImplementedError that occurs during vectoring """


# CIGAR string operation codes
CIG_ALIGN = b"M"  # alignment match
CIG_MATCH = b"="  # sequence match
CIG_SUBST = b"X"  # substitution
CIG_DELET = b"D"  # deletion
CIG_INSRT = b"I"  # insertion
CIG_SCLIP = b"S"  # soft clipping

# Regular expression pattern that matches a single CIGAR operation
# (length ≥ 1 and operation code, defined above)
CIG_PATTERN = re.compile(b"".join([rb"(\d+)([",
                                   CIG_ALIGN,
                                   CIG_MATCH,
                                   CIG_SUBST,
                                   CIG_DELET,
                                   CIG_INSRT,
                                   CIG_SCLIP,
                                   b"])"]))


def encode_base(base: int):
    if base == T_INT:
        return SUB_T
    if base == G_INT:
        return SUB_G
    if base == C_INT:
        return SUB_C
    if base == A_INT:
        return SUB_A
    raise VectorValueError(f"Invalid base: '{chr(base)}'")


def encode_compare(ref_base: int, read_base: int, read_qual: int, min_qual: int):
    return ((MATCH if ref_base == read_base
             else encode_base(read_base)) if read_qual >= min_qual
            else ANY_N ^ encode_base(ref_base))


def encode_match(read_base: int, read_qual: int, min_qual: int):
    """
    A more efficient version of encode_compare given the prior knowledge
    from the CIGAR string that the read and reference match at this
    position. Note that there is no analagous version when there is a
    known substitution because substitutions are relatively infrequent,
    so optimizing their processing would speed the program only sligtly
    while making the source code more complex and harder to maintain.
    :param read_base:
    :param read_qual:
    :param min_qual:
    :return:
    """
    return (MATCH if read_qual >= min_qual
            else ANY_N ^ encode_base(read_base))


class Indel(object):
    """
    Base class for an Insertion or Deletion (collectively, "indel")
    It is used to find alternative positions for indels by keeping track
    of an indel's current coordinates (as it is moved) and determining
    whether a specific move is valid.
    Arguments
    ---------
    rel_ins_idx: int
        The 0-indexed position of the indel with respect to the sequence
        (ref or read) with the relative insertion. This position points
        to one specific base. If the mutation is labeled an insertion,
        then the read is the sequence with the relative insertion (since
        it has a base that is not in the reference), and rel_ins_idx is
        the 0-based index of the inserted base in the coordinates of the
        read sequence. If the mutation is labeled a deletion, then the
        reference is the sequence with the relative insertion (since it
        has a base that is not in the read), and rel_ins_idx is the
        0-based index of the deleted base in the coordinates of the
        reference sequence.
    rel_del_idx (int):  The opposite of rel_ins_idx: the 0-indexed position
                        of the indel with respect to the sequence with a
                        relative deletion (that is, the read if the mutation
                        is denoted a deletion, and the ref if an insertion).
                        Because the deleted base does not actually exist
                        within the sequence whose coordinates it is based on,
                        rel_del_idx does not refer to a specific position in
                        the sequence, but instead to the two extant positions
                        in the sequence that flank the deleted position.
                        It is most convenient for the algorithm to have this
                        argument refer to the position 3' of the deleted base
                        and define the 5' position as a property.
    """

    # Define __slots__ to improve speed and memory performance.
    __slots__ = ["_ins_idx", "_ins_init", "_del_idx", "_del_init", "_tunneled"]

    # Minimum distance between an insertion and a deletion
    MIN_INDEL_DIST = 2

    def __init__(self, rel_ins_idx: int, rel_del_idx: int) -> None:
        self._ins_idx = rel_ins_idx
        self._ins_init = rel_ins_idx
        self._del_idx = rel_del_idx
        self._del_init = rel_del_idx
        self._tunneled = False

    @property
    def ins_idx(self):
        return self._ins_idx

    @property
    def del_idx5(self):
        return self._del_idx - 1

    @property
    def del_idx3(self):
        return self._del_idx

    @property
    def tunneled(self):
        return self._tunneled

    @property
    def rank(self) -> int:
        raise VectorNotImplementedError

    def reset(self):
        """ Reset the indel to its initial position, and erase its
        history of tunneling. """
        self._ins_idx = self._ins_init
        self._del_idx = self._del_init
        self._tunneled = False

    @staticmethod
    def _get_indel_by_idx(indels: list[Indel], idx: int):
        for indel in indels:
            if indel.ins_idx == idx:
                return indel

    def _peek_out_of_indel(self, indels: list[Indel], from3to5: bool):
        inc = -1 if from3to5 else 1
        idx = self.ins_idx + inc
        tunneled_indels: list[Indel] = list()
        while indel := (self._get_indel_by_idx(indels, idx)):
            idx += inc
            tunneled_indels.append(indel)
        self._tunneled = bool(tunneled_indels)
        return idx, tunneled_indels

    def _collision(self, other: Indel, swap_idx: int):
        return self.MIN_INDEL_DIST > (min(abs(swap_idx - other.del_idx5),
                                          abs(swap_idx - other.del_idx3)))

    def _collisions(self, indels: list[Indel], swap_idx: int):
        return any(self._collision(indel, swap_idx) for indel in indels)

    def step_del_idx(self, swap_idx: int):
        # Move the indel's position (self._ins_idx) to swap_idx.
        # Move self._del_idx one step in the same direction.
        if swap_idx == self.ins_idx:
            raise VectorValueError(f"swap ({swap_idx}) = ins ({self.ins_idx})")
        self._del_idx += 1 if swap_idx > self.ins_idx else -1

    def _step(self, swap_idx: int):
        self.step_del_idx(swap_idx)
        self._ins_idx = swap_idx

    @staticmethod
    def _consistent_rels(curr_rel: int, swap_rel: int):
        if curr_rel & swap_rel or (curr_rel & SUB_N
                                   and swap_rel & SUB_N):
            # Relationship between reference and read base (read_code) and
            # relationship between reference and swap base (swap_code)
            # are consistent, meaning either
            # - both match the reference
            # - one matches and the other potentially matches (i.e. low qual)
            # - one is a substitution and the other could be a substitution
            # - both are substitutions (for each code, code & SUB_N == code)
            return curr_rel
        # Otherwise, i.e.g if one base matches and the other is a substitution,
        # then the relationships are not consistent.
        return 0

    def _encode_swap(self, *args, **kwargs) -> bool:
        raise VectorNotImplementedError

    def _try_swap(self, *args, **kwargs) -> bool:
        raise VectorNotImplementedError

    def sweep(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
              min_qual: int, dels: list[Deletion], inns: list[Insertion],
              from3to5: bool, tunnel: bool):
        # Move the indel as far as possible in either the 5' or 3' direction.
        while self._try_swap(muts, ref, read, qual, min_qual, dels, inns,
                             from3to5, tunnel):
            # All actions happen in _try_swap, so loop body is empty.
            pass


class Deletion(Indel):
    @property
    def rank(self):
        return self._ins_idx

    @classmethod
    def _encode_swap(cls, ref_base: int, swap_base: int, read_base: int,
                     read_qual: int, min_qual: int):
        curr_rel = encode_compare(ref_base, read_base, read_qual, min_qual)
        swap_rel = encode_compare(swap_base, read_base, read_qual, min_qual)
        return cls._consistent_rels(curr_rel, swap_rel)

    def _swap(self, muts: bytearray, swap_idx: int, relation: int):
        """
        Arguments
        ---------
        muts: bytearray
            Mutation vector
        swap_idx: int
            Index in the region to which the deletion moves during this
            swap
        relation: int
            Relationship (match, sub, etc.) between the base located at
            swap_idx and the base in the read
        """
        # The base at swap_idx moves to self.ref_idx, so after the swap, the
        # relationship between self.ref_idx and the read base will be swap_code.
        muts[self.ins_idx] = muts[self.ins_idx] | relation
        # The base at self.ref_idx is marked as a deletion (by definition), so
        # mark the position it moves to (swap_idx) as a deletion too.
        muts[swap_idx] = muts[swap_idx] | DELET
        self._step(swap_idx)

    def _try_swap(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                  min_qual: int, dels: list[Deletion], inns: list[Insertion],
                  from3to5: bool, tunnel: bool) -> bool:
        swap_idx, tunneled_indels = self._peek_out_of_indel(dels, from3to5)
        read_idx = self.del_idx5 if from3to5 else self.del_idx3
        if (0 <= swap_idx < len(ref) and 0 <= read_idx < len(read)
                and (tunnel or not self.tunneled)
                and not self._collisions(inns, swap_idx)):
            relation = self._encode_swap(ref[self.ins_idx], ref[swap_idx],
                                         read[read_idx], qual[read_idx],
                                         min_qual)
            if relation:
                self._swap(muts, swap_idx, relation)
                for indel in tunneled_indels:
                    indel.step_del_idx(swap_idx)
                return True
        return False


class Insertion(Indel):
    @property
    def rank(self):
        return self._del_idx

    def stamp(self, muts: bytearray):
        if 0 <= self.del_idx5 < len(muts):
            muts[self.del_idx5] = muts[self.del_idx5] | INS_5
        if 0 <= self.del_idx3 < len(muts):
            muts[self.del_idx3] = muts[self.del_idx3] | INS_3

    @classmethod
    def _encode_swap(cls, ref_base: int, read_base: int, read_qual: int,
                     swap_base: int, swap_qual: int, min_qual: int):
        curr_rel = encode_compare(ref_base, read_base, read_qual, min_qual)
        swap_rel = encode_compare(ref_base, swap_base, swap_qual, min_qual)
        return cls._consistent_rels(curr_rel, swap_rel)

    def _swap(self, muts: bytearray, ref_idx: int,
              swap_idx: int, relation: int):
        """
        Arguments
        mut_vectors (bytearray): mutation vector
        swap_idx (int): the index in the read to which the deletion moves
                        during this swap
        swap_code (int): the relationship (match, sub, etc.) between the
                         base located at swap_idx and the base in the ref
        """
        # The base at ins_idx moves to swap_idx, so after the swap, the
        # relationship between ref_idx and the read base will be relation.
        muts[ref_idx] = muts[ref_idx] | relation
        self._step(swap_idx)
        # Mark the new positions of the insertion.
        self.stamp(muts)

    def _try_swap(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                  min_qual: int, dels: list[Deletion], inns: list[Insertion],
                  from3to5: bool, tunnel: bool) -> bool:
        swap_idx, tunneled_indels = self._peek_out_of_indel(inns, from3to5)
        ref_idx = self.del_idx5 if from3to5 else self.del_idx3
        if (0 <= swap_idx < len(read) and 0 <= ref_idx < len(ref)
                and (tunnel or not self.tunneled)
                and not self._collisions(dels, swap_idx)):
            relation = self._encode_swap(ref[ref_idx], read[self.ins_idx],
                                         qual[self.ins_idx], read[swap_idx],
                                         qual[swap_idx], min_qual)
            if relation:
                self._swap(muts, ref_idx, swap_idx, relation)
                for indel in tunneled_indels:
                    indel.step_del_idx(swap_idx)
                return True
        return False


def sweep_indels(muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                 min_qual: int, dels: list[Deletion], inns: list[Insertion],
                 from3to5: bool, tunnel: bool):
    """
    For every insertion and deletion,
    Parameters
    ----------
    muts: bytearray
        Mutation vector
    ref: bytes
        Sequence of the region of the reference that the mutation vector
        covers
    read: bytes
        Sequence of the read
    qual: bytes
        Phred quality scores of the read, encoded as ASCII characters
    min_qual: int
        The minimum Phred quality score needed to consider a base call
        informative: integer value of the ASCII character
    dels: list[Deletion]
        List of deletions identified by ```vectorize_read```
    inns: list[Insertion]
        List of insertions identified by ```vectorize_read```
    from3to5: bool
        Whether to move indels in the 3' -> 5' direction (True) or the
        5' -> 3' direction (False)
    tunnel: bool
        Whether to allow tunneling
    """
    # Collect all indels into one list.
    indels: list[Indel] = list()
    indels.extend(dels)
    indels.extend(inns)
    # Reset each indel to its initial state. This operation does nothing
    # the first time sweep_indels is called because all indels start in
    # their initial state (by definition). But the indels may move when
    # this function runs, so resetting is necessary at the beginning of
    # the second and subsequent calls to sweep_indels to ensure that the
    # algorithm starts from the initial state every time.
    for indel in indels:
        indel.reset()
    # Sort the indels by their rank, which is
    sort_rev = from3to5 != tunnel
    indels.sort(key=lambda idl: idl.rank, reverse=sort_rev)
    while indels:
        indel = indels.pop()
        indel.sweep(muts, ref, read, qual, min_qual,
                    dels, inns, from3to5, tunnel)
        i = len(indels)
        if sort_rev:
            while i > 0 and indel.rank > indels[i - 1].rank:
                i -= 1
        else:
            while i > 0 and indel.rank < indels[i - 1].rank:
                i -= 1
        if i < len(indels):
            indels.insert(i, indel)


def get_ambids(muts: bytearray, ref: bytes, read: bytes, qual: bytes,
               min_qual: int, dels: list[Deletion], inns: list[Insertion]):
    """
    Find and label all positions in the vector that are ambiguous due to
    insertions and deletions.
    Parameters
    ----------
    muts: bytearray
        Mutation vector
    ref: bytes
        Sequence of the region of the reference that the mutation vector
        covers
    read: bytes
        Sequence of the read
    qual: bytes
        Phred quality scores of the read, encoded as ASCII characters
    min_qual: int
        The minimum Phred quality score needed to consider a base call
        informative: integer value of the ASCII character
    dels: list[Deletion]
        List of deletions identified by ```vectorize_read```
    inns: list[Insertion]
        List of insertions identified by ```vectorize_read```
    """
    # Each indel might be able to be moved in the 5' -> 3' direction
    # (from3to5 is False) or 3' -> 5' direction (from3to5 is True).
    # Test both directions.
    for from3to5 in (False, True):
        # For each indel, try to move it as far as it can go in the
        # direction indicated by from3to5. Allow tunneling so that any
        # runs of consecutive insertions or consecutive deletions can
        # effectively move together.
        sweep_indels(muts, ref, read, qual, min_qual,
                     dels, inns, from3to5, tunnel=True)
        if any(d.tunneled for d in dels) or any(i.tunneled for i in inns):
            # If any indel tunneled,
            sweep_indels(muts, ref, read, qual, min_qual,
                         dels, inns, from3to5, tunnel=False)


def parse_cigar(cigar_string: bytes):
    """
    Yield the fields of a CIGAR string as pairs of (operation, length),
    where operation is 1 byte indicating the CIGAR operation and length
    is a positive integer indicating the number of bases from the read
    that the operation consumes. Note that in the CIGAR string itself,
    each length precedes its corresponding operation.
    Parameters
    ----------
    cigar_string: bytes
        CIGAR string from a SAM file. For full documentation, refer to
        https://samtools.github.io/hts-specs/
    Yield
    -----
    bytes (length = 1)
        Current CIGAR operation
    int (≥ 1)
        Length of current CIGAR operation
    Examples
    --------
    >>> list(parse_cigar(b"17=1X43=1D26="))
    [(b'=', 17), (b'X', 1), (b'=', 43), (b'D', 1), (b'=', 26)]
    """
    # Length-0 CIGAR strings are forbidden.
    if not cigar_string:
        raise VectorValueError("CIGAR string is empty")
    # If the CIGAR string has any invalid bytes (e.g. an unrecognized
    # operation byte, an operation longer than 1 byte, a length that is
    # not a positive integer, or any extraneous characters), then the
    # regular expression parser will simply skip these invalid bytes.
    # In order to catch such problems, keep track of the number of
    # bytes matched from the CIGAR string. After reading the CIGAR, if
    # the number of bytes matched is smaller than the length of the
    # CIGAR string, then some bytes must have been skipped, indicating
    # that the CIGAR string contained at least one invalid byte.
    num_bytes_matched = 0
    # Find every operation in the CIGAR string that matches the regular
    # expression.
    for match in CIG_PATTERN.finditer(cigar_string):
        length_bytes, operation = match.groups()
        # Convert the length field from bytes to int and verify that it
        # is a positive integer.
        if (length_int := int(length_bytes)) < 1:
            raise VectorValueError("length of CIGAR operation must be ≥ 1")
        # Add the total number of bytes in the current operation to the
        # count of the number of bytes matched from the CIGAR string.
        num_bytes_matched += len(length_bytes) + len(operation)
        # Note that the fields are yielded as (operation, length), but
        # in the CIGAR string itself, the order is (length, operation).
        yield operation, length_int
    # Confirm that all bytes in the CIGAR string were matched by the
    # regular expression. Note: This check will only be performed if
    # the entire CIGAR string is read. Thus, it is essential to read
    # the entire CIGAR string, even if the read extends beyond the
    # region for which the mutation vector is being computed.
    if num_bytes_matched != len(cigar_string):
        raise VectorValueError(f"Invalid CIGAR: '{cigar_string.decode()}'")


def op_consumes_ref(op: bytes) -> bool:
    """ Return whether the CIGAR operation consumes the reference. """
    return op != CIG_INSRT and op != CIG_SCLIP


def op_consumes_read(op: bytes) -> bool:
    """ Return whether the CIGAR operation consumes the read. """
    return op != CIG_DELET


class SamFlag(object):
    """ Represents the set of 12 boolean flags for a SAM record. """

    # Define __slots__ to improve speed and memory performance.
    __slots__ = ["paired", "proper", "unmap", "munmap", "rev", "mrev",
                 "first", "second", "secondary", "qcfail", "dup", "supp"]

    # Maximum value of a valid SAM flag representation, corresponding
    # to all 12 flags set to 1: 111111111111 (binary) = 4095 (decimal)
    MAX_FLAG: int = 2 ** len(__slots__) - 1
    # Pattern for padding the left of the binary string with 0s.
    PATTERN = "".join(["{:0>", str(len(__slots__)), "}"])

    def __init__(self, flag: int):
        """
        Validate the integer value of the SAM flag, then set the 12
        individual flag values. To maximize speed, all flags are set in
        a single one-line operation, each step of which is explained:
        1.  Convert the flag (int) to a binary representation (str) that
            starts with '0b':
            >>> (flag_bin := bin(flag_int := 83))
            '0b1010011'
        2.  Remove the prefix '0b':
            >>> (flag_bits := flag_bin[2:])
            '1010011'
        3.  Pad the left side of the string with 0 up to a length of 12:
            >>> (PATTERN := "".join(["{:0>", str(num_flags := 12), "}"]))
            '{:0>12}'
            >>> (all_bits := PATTERN.format(flag_bits))
            '000001010011'
        4.  Convert '1' to True and '0' to False, and assign to the 12
            flag bits in order from greatest (supp: 2048) to least
            (paired: 1) numerical value of the flag bit:
            >>> (supp, dup, qcfail, secondary, second, first, mrev, rev,
            ...  munmap, unmap, proper, paired) = map(bool, map(int, all_bits))
            >>> paired, rev, second, qcfail
            (True, True, False, False)
        Parameters
        ----------
        flag: int
            The integer value of the SAM flag. For documentation, see
            https://samtools.github.io/hts-specs/
        Examples
        --------
        >>> flag0099 = SamFlag(99)
        >>> flag0099.paired, flag0099.rev
        (True, False)
        """
        if not 0 <= flag <= self.MAX_FLAG:
            raise VectorValueError(f"Invalid flag: '{flag}'")
        (self.supp, self.dup, self.qcfail, self.secondary,
         self.second, self.first, self.mrev, self.rev,
         self.munmap, self.unmap, self.proper, self.paired) = (
            x == '1' for x in self.PATTERN.format(bin(flag)[2:])
        )


class SamRead(object):
    # Define __slots__ to improve speed and memory performance.
    __slots__ = ["qname", "flag", "paired", "rname", "pos", "mapq", "cigar",
                 "tlen", "seq", "qual", "min_qual"]

    # Minimum number of fields in a valid SAM record
    MIN_FIELDS = 11

    def __init__(self, line: bytes):
        fields = line.rstrip().split(b"\t")
        if len(fields) < self.MIN_FIELDS:
            raise VectorValueError(f"Invalid SAM line:\n{line}")
        self.qname = fields[0]
        self.flag = SamFlag(int(fields[1]))
        self.paired = self.flag.paired
        self.rname = fields[2].decode()
        self.pos = int(fields[3])
        self.mapq = int(fields[4])
        self.cigar = fields[5]
        # RNEXT, PNEXT, and TLEN are not used during vectoring,
        # so they are commented out to reduce processing time
        # but still shown here in case they are ever needed.
        # self.rnext = fields[6]
        # self.pnext = int(fields[7])
        # self.tlen = int(fields[8])
        self.seq = fields[9]
        self.qual = fields[10]
        if len(self.seq) != len(self.qual):
            raise VectorValueError(f"Lengths of seq ({len(self.seq)}) and qual "
                                   f"string {len(self.qual)} did not match.")


def vectorize_read(read: SamRead,
                   muts: bytearray,
                   region_seq: bytes,
                   region_length: int,
                   region_end5: int,
                   min_qual: int,
                   ambid: bool):
    """
    Generate and return a mutation vector of an aligned read over a
    given region of the reference sequence.
    Parameters
    ----------
    read: SamRead
        Read from SAM file to be vectorized
    muts: bytearray
        Mutation vector (initially blank) into which to write bytes
    region_seq: bytes
        Sequence of the region for which to compute the mutation vector
        (only the region, not any other part of the reference sequence)
    region_length: int (≥ region_end5)
        3'-most coordinate of the region with respect to the entire
        reference sequence (1-indexed, includes coordinate)
    region_end5: int (≥ 1)
        5'-most coordinate of the region with respect to the entire
        reference sequence (1-indexed, includes coordinate)
    min_qual: int
        ASCII encoding of the minimum Phred score to accept a base call
    ambid: bool
        Whether to find and label all ambiguous insertions and deletions
    Return
    ------
    bytearray
        Mutation vector, whose length either equals that of the region
        or is zero to indicate an error occurred during vectorization.
    """
    if region_length != len(muts):
        raise VectorValueError(f"Region is {region_length} nt, but mutation "
                               f"vector is {len(muts)} nt.")
    if region_length != len(region_seq):
        raise VectorValueError(f"Region is {region_length} nt, but sequence "
                               f"is {len(region_seq)} nt.")
    # Indexes of the 5' and 3' ends of the current CIGAR operation with
    # respect to the read; 0-indexed, using Python half-open intervals
    read_idx5 = 0
    read_idx3 = 0
    # Indexes of the 5' and 3' ends of the current CIGAR operation with
    # respect to the region; 0-indexed, using Python half-open intervals
    region_idx5 = read.pos - region_end5
    region_idx3 = region_idx5
    # Number of bases truncated from the end of the operation
    truncated = 0
    # Record all deletions and insertions.
    dels: list[Deletion] = list()
    inns: list[Insertion] = list()
    # Read the CIGAR string one operation at a time.
    for cigar_op, op_length in parse_cigar(read.cigar):
        # Update the coordinates, with respect to the region and read,
        # that correspond to the 3' end of the current CIGAR operation.
        if op_consumes_ref(cigar_op):
            region_idx3 += op_length
        if op_consumes_read(cigar_op):
            read_idx3 += op_length
        # Check if the part of the read that corresponds to the current
        # CIGAR operation at all overlaps the region of interest.
        # Note: This loop does not terminate when the CIGAR operation
        # exits the region of interest, even though no more matches or
        # mutations will be appended to the end of mut_vectors afterwards.
        # This behavior forces the entire CIGAR string to be read, which
        # is necessary for parse_cigar to validate the CIGAR string.
        if region_idx3 > 0 and region_idx5 < region_length:
            if region_idx5 < 0:
                # If the 3' end of the CIGAR operation overlaps the
                # region but the 5' end does not, then truncate the
                # 5' end of the current CIGAR operation so that the
                # operation starts at the 5' end of the region.
                # Decrease the length of the CIGAR operation because it
                # is being truncated. Note: region_idx5 < 0.
                op_length += region_idx5
                if op_consumes_read(cigar_op):
                    # If the CIGAR operation consumes the read, advance
                    # the index of the 5' end of the CIGAR operation
                    # with respect to the read. Note: region_idx5 < 0.
                    read_idx5 -= region_idx5
                # Advance the index of the 5' end of the CIGAR operation
                # with respect to the region to 0; that is, the CIGAR
                # operation now starts at the 5' end of the region.
                region_idx5 = 0
            if region_idx3 > region_length:
                # If the 5' end of the CIGAR operation overlaps the
                # region but the 3' end does not, then truncate the
                # 3' end of the current CIGAR operation so that the
                # operation ends at the 3' end of the region.
                # First, find the number of positions to truncate from
                # the 3' end; truncated is guaranteed to be > 0.
                truncated = region_idx3 - region_length
                # Decrease the length of the CIGAR operation because it
                # is being truncated.
                op_length -= truncated
                if op_consumes_read(cigar_op):
                    # If the CIGAR operation consumes the read, reduce
                    # the index of the 3' end of the CIGAR operation
                    # with respect to the read.
                    read_idx3 -= truncated
                # Reduce the index of the 3' end of the CIGAR operation
                # with respect to the region to equal the length of the
                # region; that is, the CIGAR operation now ends at the
                # 3' end of the region.
                region_idx3 = region_length
            # Act based on the CIGAR operation and its length.
            if cigar_op == CIG_MATCH:
                # The read and reference sequences match over the entire
                # CIGAR operation.
                for base, qual in zip(read.seq[read_idx5: read_idx3],
                                      read.qual[read_idx5: read_idx3]):
                    muts[region_idx5] = encode_match(base, qual, min_qual)
                    region_idx5 += 1
            elif cigar_op == CIG_ALIGN or cigar_op == CIG_SUBST:
                # The read contains only matches or substitutions (no
                # indels) relative to the reference over the entire
                # CIGAR operation.
                for ref_base, read_base, read_qual in zip(
                        region_seq[region_idx5: region_idx3],
                        read.seq[read_idx5: read_idx3],
                        read.qual[read_idx5: read_idx3],
                        strict=True):
                    muts[region_idx5] = encode_compare(ref_base, read_base,
                                                       read_qual, min_qual)
                    region_idx5 += 1
            elif cigar_op == CIG_DELET:
                # The portion of the reference sequence corresponding
                # to the CIGAR operation is deleted from the read.
                # Create one Deletion object for each base in the
                # reference sequence that is missing from the read.
                while region_idx5 < region_idx3:
                    dels.append(Deletion(region_idx5, read_idx5))
                    muts[region_idx5] = DELET
                    region_idx5 += 1
            elif cigar_op == CIG_INSRT:
                # The read contains an insertion of one or more bases
                # that are not present in the reference sequence.
                # Create one Insertion object for each base in the read
                # sequence that is not present in the reference. Every
                # mutation needs to be assigned a coordinate in the
                # region in order to appear at that coordinate in the
                # mutation vector. But each inserted base, being absent
                # from the reference, does not correspond to a single
                # coordinate in the region; instead, each inserted base
                # lies between two coordinates in the region. Either of
                # these coordinates could be chosen; this code assigns
                # the 3' coordinate to the insertion. For example, if
                # two bases are inserted between coordinates 45 and 46
                # of the region, then both will be given coordinate 46.
                # The reason for this convention is that the math is
                # simpler than it would be if using the 5' coordinate.
                # Because region_idx5 is, by definition, immediately 3'
                # of the previous CIGAR operation; and the bases that
                # are inserted lie between the previous and subsequent
                # CIGAR operations; region_idx5 is the coordinate
                # immediately 3' of the inserted bases. In this special
                # case, region_idx5 also equals region_idx3 (because
                # the insertion does not consume the reference, so
                # region_idx3 += op_length was not run at the beginning
                # of this loop iteration), as well as the length of mut_vectors
                # (because Python is 0-indexed, so the length of a range
                # of indexes such as [0, 1, ... , 45] equals the value
                # of the next index in the range, 46). Thus, there are
                # three variables that already equal the 3' coordinate
                # and none that equal the 5' coordinate.
                while read_idx5 < read_idx3:
                    inns.append(Insertion(read_idx5, region_idx5))
                    read_idx5 += 1
                # Insertions do not consume the reference, so do not add
                # any information to mut_vectors yet; it will be added later.
            elif cigar_op == CIG_SCLIP:
                # Bases were soft-clipped from the 5' or 3' end of the
                # read during alignment. Like insertions, they consume
                # the read but not the reference. Unlike insertions,
                # they are not mutations, so they do not require any
                # additional processing.
                pass
            else:
                raise VectorValueError(
                    f"Invalid CIGAR operation: '{cigar_op.decode()}'")
        if truncated:
            # If the current operation was truncated because it extended
            # past the 3' end of the region, then the 3' ends of the
            # region and read need to be reset to their values before
            # truncation. Otherwise, all subsequent CIGAR operations
            # will start and end 5' of where they should.
            region_idx3 += truncated
            if op_consumes_read(cigar_op):
                read_idx3 += truncated
            truncated = 0
        # Advance the 5' positions in the region and read to the current
        # 3' positions so that the next CIGAR operation lies immediately
        # 3' of the current operation. The 3' positions will be advanced
        # at the beginning of the next iteration (if any) of the loop.
        region_idx5 = region_idx3
        read_idx5 = read_idx3
    if len(muts) != region_length:
        raise VectorValueError(f"Mutation vector is {len(muts)} nt, "
                               f"but region is {region_length} nt.")
    # Verify that the sum of all CIGAR operations that consumed the read
    # equals the length of the read. The former equals read_idx5 because
    # for each operation that consumed the read, the length of the
    if read_idx5 != len(read.seq):
        raise VectorValueError(
            f"CIGAR string '{read.cigar.decode()}' consumed {read_idx5} "
            f"bases from read, but read is {len(read.seq)} bases long.")
    # Add insertions to mut_vectors.
    for ins in inns:
        ins.stamp(muts)
    # Label all positions that are ambiguous due to indels.
    if ambid and (dels or inns):
        get_ambids(muts, region_seq, read.seq, read.qual, min_qual, dels, inns)


def vectorize_line(line: bytes, muts: bytearray, seq: bytes,
                   length: int, end5: int, ref: str, qmin: int, ambid: bool):
    read = SamRead(line)
    if read.rname != ref:
        raise VectorValueError(f"Read '{read.qname.decode()}' had reference "
                               f"'{read.rname}' (≠ '{ref}')")
    vectorize_read(read, muts, seq, length, end5, qmin, ambid)


def vectorize_pair(line1: bytes, line2: bytes, muts: bytearray, seq: bytes,
                   length: int, end5: int, ref: str, qmin: int, ambid: bool):
    # Parse lines 1 and 2 into SAM reads.
    read1 = SamRead(line1)
    read2 = SamRead(line2)
    # Ensure that reads 1 and 2 are compatible mates.
    if not read1.flag.paired:
        raise VectorValueError(f"Read 1 ({read1.qname.decode()}) was not "
                               f"paired, but read 2 ('{read2.qname.decode()}') "
                               f"was given")
    if not read2.flag.paired:
        raise VectorValueError(f"Read 2 ({read2.qname.decode()}) was not "
                               f"paired, but read 1 ({read1.qname.decode()}) "
                               f"was given")
    if read1.qname != read2.qname:
        raise VectorValueError(f"Reads 1 ({read1.qname.decode()}) and 2 "
                               f"({read2.qname.decode()}) had different names")
    if read1.rname != read2.rname:
        raise VectorValueError(f"Read '{read1.qname.decode()}' had "
                               "different references for mates 1 "
                               f"('{read1.rname}') and 2 "
                               f"('{read2.rname}')")
    if read1.rname != ref:
        raise VectorValueError(f"Read '{read1.qname.decode()}' had reference "
                               f"'{read1.rname}' (≠ '{ref}')")
    if not (read1.flag.first and read2.flag.second):
        raise VectorValueError(f"Read '{read1.qname.decode()}' had mate 1 "
                               f"labeled {2 - read1.flag.first} and mate 2 "
                               f"labeled {1 + read2.flag.second}")
    if read1.flag.rev == read2.flag.rev:
        raise VectorValueError(f"Read '{read1.qname.decode()}' had "
                               "mates 1 and 2 facing the same way")
    # Vectorize read 1.
    muts1 = muts.copy()
    vectorize_read(read1, muts1, seq, length, end5, qmin, ambid)
    # Vectorize read 2.
    muts2 = muts.copy()
    vectorize_read(read2, muts2, seq, length, end5, qmin, ambid)
    # Compute the consensus of reads 1 and 2.
    for i, (m1, m2) in enumerate(zip(muts1, muts2, strict=True)):
        muts[i] = inter if (inter := m1 & m2) else m1 | m2



from logging import getLogger
from pathlib import Path
import re
from typing import BinaryIO

from ..util import path
from ..util.shell import run_cmd, SAMTOOLS_CMD

logger = getLogger(__name__)

# SAM file format specifications
SAM_HEADER = b"@"
SAM_DELIMITER = b"\t"
SAM_ALIGN_SCORE = b"AS:i:"
SAM_EXTRA_SCORE = b"XS:i:"


def index_bam(bam: Path, n_padd: int = 0):
    """ Build an index of a BAM file using ```samtools index```. """
    logger.info(f"Began building BAM index of {bam}")
    cmd = [SAMTOOLS_CMD, "index", "-@", n_padd - 1, bam]
    run_cmd(cmd)
    logger.info(f"Ended building BAM index of {bam}: "
                f"{bam.with_suffix(path.BAI_EXT)}")


def sort_xam(xam_inp: Path, xam_out: Path, *,
             name: bool = False, n_padd: int = 0):
    """ Sort a SAM or BAM file using ```samtools sort```. """
    logger.info(f"Began sorting {xam_inp}")
    cmd = [SAMTOOLS_CMD, "sort", "-@", n_padd]
    if name:
        # Sort by name instead of coordinate.
        cmd.append("-n")
    cmd.extend(["-o", xam_out, xam_inp])
    # Make the output directory.
    xam_out.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensured directory: {xam_out.parent}")
    run_cmd(cmd)
    logger.info(f"Ended sorting {xam_inp} to {xam_out}")


def view_xam(xam_inp: Path,
             xam_out: Path, *,
             ref: str | None = None,
             end5: int | None = None,
             end3: int | None = None,
             n_padd: int = 0):
    """ Convert between SAM and BAM formats, or extract reads aligning
    to a specific reference/section using ```samtools view```. """
    logger.info(f"Began viewing {xam_inp}")
    cmd = [SAMTOOLS_CMD, "view", "-@", n_padd, "-h"]
    if xam_out.suffix == path.BAM_EXT:
        # Write a binary (BAM) file.
        cmd.append("-b")
    # Input and output files
    cmd.extend(("-o", xam_out, xam_inp))
    # Reference and section specification
    if ref is not None:
        if end5 is not None and end3 is not None:
            # View only reads aligning to a section of this reference.
            cmd.append(f"{ref}:{end5}-{end3}")
        else:
            # View only reads aligning to this reference.
            cmd.append(ref)
            if end5 is not None:
                logger.warning(f"Got end5 = {end5} but not end3")
            if end3 is not None:
                logger.warning(f"Got end3 = {end3} but not end5")
    elif end5 is not None or end5 is not None:
        logger.warning(f"Options end5 and end3 require a reference name")
    # Make the output directory.
    xam_out.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensured directory: {xam_out.parent}")
    # Run the command.
    run_cmd(cmd)
    logger.info(f"Ended viewing {xam_inp} as {xam_out}")


def dedup_sam(sam_inp: Path, sam_out: Path):
    """ Remove SAM reads that map equally to multiple locations. """

    logger.info(f"Began deduplicating {sam_inp}")

    pattern_a = re.compile(SAM_ALIGN_SCORE + rb"(\d+)")
    pattern_x = re.compile(SAM_EXTRA_SCORE + rb"(\d+)")

    min_fields = 11
    max_flag = 4095  # 2^12 - 1

    def get_score(line: bytes, ptn: re.Pattern[bytes]):
        return (float(match.groups()[0])
                if (match := ptn.search(line)) else None)

    def is_best_align(line: bytes):
        return ((score_x := get_score(line, pattern_x)) is None
                or score_x < get_score(line, pattern_a))

    def read_is_paired(line: bytes):
        info = line.split()
        if len(info) < min_fields:
            raise ValueError(f"Invalid SAM line:\n{line.decode()}")
        flag = int(info[1])
        if flag < 0 or flag > max_flag:
            raise ValueError(f"Invalid SAM flag: {flag}")
        return bool(flag % 2)

    def write_summary_single(written: int, skipped: int):
        total = written + skipped
        f_written = (round(100 * written / total, 3) if total
                     else float('nan'))
        return (f"Summary of removing multiply-mapped reads from {sam_inp}\n"
                f"Total number of single-end reads: {total}\n"
                f"Uniquely-mapped reads (written):  {written}\n"
                f"Multiply-mapped reads (skipped):  {skipped}\n"
                f"Percent uniquely mapped (written): {f_written} %")

    def write_summary_paired(pairs_written: int, pairs_skipped: int,
                             mates_written: int, mates_skipped: int):
        mates = mates_written + mates_skipped
        pairs = pairs_written + pairs_skipped
        pairs_one_mate = 2 * pairs - mates
        pairs_both_mates = pairs - pairs_one_mate
        f_written = (round(100 * pairs_written / pairs, 3) if pairs
                     else float('nan'))
        return (f"Summary of removing multiply-mapped reads from {sam_inp}\n"
                f"Total number of mates aligning: {mates}\n"
                f"Total number of pairs of mates: {pairs}\n"
                f"Pairs with both mates aligning: {pairs_both_mates}"
                f"Pairs with a mate not aligning: {pairs_one_mate}"
                f"Uniquely-mapped pairs (written):  {pairs_written}\n"
                f"Multiply-mapped pairs (skipped):  {pairs_skipped}\n"
                f"Percent uniquely mapped (written): {f_written} %")

    def iter_single(sam: BinaryIO, line: bytes):
        """ For each read, yield the best-scoring alignment, excluding
        reads that aligned equally well to multiple locations. """
        n_copy = 0
        n_skip = 0
        while line:
            if is_best_align(line):
                n_copy += 1
                yield line
            else:
                n_skip += 1
            line = sam.readline()
        logger.debug(write_summary_single(n_copy, n_skip))

    def iter_paired(sam: BinaryIO, line1: bytes):
        """ For each pair of reads, yield the pair of alignments for
        which both the forward alignment and the reverse alignment in
        the pair scored best among all alignments for the forward and
        reverse reads, respectively. Exclude pairs for which the forward
        and/or reverse read aligned equally well to multiple locations,
        or for which the best alignments for the forward and reverse
        reads individually are not part of the same alignment pair. """
        n_pairs_written = 0
        n_pairs_skipped = 0
        n_mates_written = 0
        n_mates_skipped = 0
        while line1:
            if SAM_DELIMITER not in line1:
                # If the line is blank, skip it.
                line1 = sam.readline()
                continue
            # Check if there is at least one more line in the file.
            if (line2 := sam.readline()) and SAM_DELIMITER in line2:
                # Check if the reads on lines 1 and 2 are mates.
                if (line1.split(SAM_DELIMITER, 1)[0]
                        == line2.split(SAM_DELIMITER, 1)[0]):
                    # If they are mates of each other, check if the
                    # lines represent the best alignment for both.
                    if is_best_align(line1) and is_best_align(line2):
                        # If so, then yield both mates.
                        n_pairs_written += 1
                        n_mates_written += 2
                        yield line1
                        yield line2
                    else:
                        # Otherwise, skip both mates.
                        n_pairs_skipped += 1
                        n_mates_skipped += 2
                    # Read the next line from the file.
                    line1 = sam.readline()
                    continue
            # If line2 does not have the mate of line1, then check only
            # line1 now and consider line2 on the next iteration.
            if is_best_align(line1):
                # Yield only line1.
                n_pairs_written += 1
                n_mates_written += 1
                yield line1
            else:
                # Skip line1.
                n_pairs_skipped += 1
                n_mates_skipped += 1
            # Since line2 was not used yet, consider it on the next
            # iteration instead of reading another line from the file.
            line1 = line2
        logger.debug(write_summary_paired(n_pairs_written, n_pairs_skipped,
                                          n_mates_written, n_mates_skipped))

    # Make the output directory.
    sam_out.parent.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Ensuring directory: {sam_out.parent}")

    # Deduplicate the alignments.
    with (open(sam_inp, "rb") as sami, open(sam_out, "xb") as samo):
        # Copy the entire header from the input to the output SAM file.
        n_header = 0
        while (line_ := sami.readline()).startswith(SAM_HEADER):
            samo.write(line_)
            n_header += 1
        logger.debug(
            f"Copied {n_header} header lines from {sam_inp} to {sam_out}")
        # Copy only the reads that mapped best to one location from the
        # input to the output SAM file.
        if line_:
            is_paired = read_is_paired(line_)
            # Determine whether the reads in the SAM file are paired.
            if is_paired:
                endedness = "paired"
                lines = iter_paired(sami, line_)
            else:
                endedness = "single"
                lines = iter_single(sami, line_)
            logger.debug(f"Deduplicating {endedness}-end SAM file: {sam_inp}")
            # Write the selected lines to the output SAM file.
            for line_ in lines:
                samo.write(line_)
        else:
            logger.critical(f"SAM file {sam_inp} contained no reads")

    logger.info(f"Ended deduplicating {sam_inp} to {sam_out}")


'''
primer1 = "CAGCACTCAGAGCTAATACGACTCACTATA"
primer1rc = "TATAGTGAGTCGTATTAGCTCTGAGTGCTG"
primer2 = "TGAAGAGCTGGAACGCTTCACTGA"
primer2rc = "TCAGTGAAGCGTTCCAGCTCTTCA"
adapters5 = (primer1, primer2rc)
adapters3 = (primer2, primer1rc)
'''



import os
import subprocess
import yaml


def split(software):
    software = software.strip()
    software = software.replace(' ', '').replace(',', '')
    for i, c in enumerate(software):
        if c in ('<', '>', '=', '!', '~'):
            return software[:i], software[i:]
    raise ValueError('No version condition found in {}'.format(software))
        

def split_conditions(conditions):
    """
    example:
    '>=3.4.5!=3.5.0<4.0' -> [('>=', '3.4.5'), ('!=', '3.5.0'), ('<', '4.0')]
    """
    out = []
    i = 0
    while i < len(conditions):
        for operator in ('>=', '>', '<=', '<', '!=', '='):
            if conditions.startswith(operator, i):
                i += len(operator)
                j = i
                while j < len(conditions) and not conditions[j] in ('>=', '>', '<=', '<', '!=', '=','!'):
                    j += 1
                out.append((operator, conditions[i:j]))
                i = j
                break
        else:
            raise ValueError('Invalid operator in conditions')
    return out


def is_number_or_dot(s):
    for c in s:
        if not c in ('0','1','2','3','4','5','6','7','8','9','.'):
            return False
    return True
    
def compare_version_numbers(a, b):
    """
    Example:
    '1.2.3', '1.2.3' -> 0
    '1.2.3', '1.2.4' -> -1
    '1.2.4', '1.2.3' -> 1
    """
    
    # compare numbers
    a = a.split('.')
    b = b.split('.')
    # make sure they are the same length
    while len(a) < len(b):
        a.append('0')
    while len(b) < len(a):
        b.append('0')
    a, b = ''.join(a), ''.join(b)
    if a == b:
        return 0
    return (a > b) - (a < b)

def validate_condition(installed_version, required_version, condition):
    if condition == '=':
        return installed_version == required_version
    if condition == '>':
        return compare_version_numbers(installed_version, required_version) > 0
    if condition == '<':
        return compare_version_numbers(installed_version, required_version) < 0
    if condition == '>=':
        return compare_version_numbers(installed_version, required_version) >= 0
    if condition == '<=':
        return compare_version_numbers(installed_version, required_version) <= 0
    if condition == '!=':
        return installed_version != required_version
    raise Exception('condition {} not recognized'.format(condition))

def validate_all_conditions(installed_version, dependencies):
    """Example:
    ('fastqc', '0.11.8', [('>=', '0.11.8'), ('!=', '0.11.9')]) -> True
    ('fastqc', '0.11.9', [('>=', '0.11.8'), ('!=', '0.11.9')]) -> False
    """
    for condition, required_version in dependencies:
        if not validate_condition(installed_version, required_version, condition):
            return False
    return True
        

dependencies_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'env.yml')

dependencies_yml = yaml.load(open(dependencies_file), Loader=yaml.FullLoader)['dependencies']
del dependencies_yml[-1]

dependencies_txt = {split(v)[0]:split(v)[1] for v in dependencies_yml}
dependencies = {k:split_conditions(split(v)[1]) for k, v in dependencies_txt.items()}
requirements = open(os.path.join(os.path.dirname(os.path.dirname(__file__)), 'requirements.txt')).readlines()
dependencies_txt = {**dependencies_txt, **{split(v)[0]:split(v)[1] for v in requirements}}
use_requirements = ['cutadapt']
requirements = {split(v)[0]:split_conditions(split(v)[1]) for v in requirements if split(v)[0] in use_requirements}
dependencies = {**dependencies, **requirements}

dependencies_installation_page = 'https://rouskinlab.github.io/dreem/dreem/dependencies.html'

def run_cmd(cmd):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    output, error = process.communicate()
    return output.decode('utf-8').strip()

def check_fastqc_exists():
    version = dependencies['fastqc']
    version_txt = dependencies_txt['fastqc']
    fastqc_installed_version = run_cmd(['fastqc', '--version'])
    assert fastqc_installed_version.startswith('FastQC'), 'fastqc is not installed'
    fastqc_installed_version = fastqc_installed_version.split('\n')[0].split(' ')[1][1:]
    assert validate_all_conditions(fastqc_installed_version, version), 'fastqc version is not correct: {} incompatible with {}. Get installation instructions here: {}'.format(fastqc_installed_version, version_txt, dependencies_installation_page)
    
def check_cutadapt_exists():
    version, version_txt = dependencies['cutadapt'], dependencies_txt['cutadapt']
    cutadapt_installed_version = run_cmd(['cutadapt', '--version'])
    assert not 'cutadapt: not found' in cutadapt_installed_version, 'cutadapt is not installed'
    assert validate_all_conditions(cutadapt_installed_version, version), 'cutadapt version is not correct: {} incompatible with {}. Get installation instructions here: {}'.format(cutadapt_installed_version, version_txt, dependencies_installation_page)
    
def check_bowtie2_exists():
    version, version_txt = dependencies['bowtie2'], dependencies_txt['bowtie2']
    bowtie2_installed_version = run_cmd(['bowtie2', '--version']).split('\n')[0].split(' ')
    assert bowtie2_installed_version[1] == 'version', 'bowtie2 is not installed'
    assert validate_all_conditions(bowtie2_installed_version[2], version), 'bowtie2 version is not correct {} incompatible with {}. Get installation instructions here: {}'.format(bowtie2_installed_version[2], version_txt, dependencies_installation_page)
    
def check_samtools_exists():
    version, version_txt = dependencies['samtools'], dependencies_txt['samtools']
    samtools_installed_version = run_cmd(['samtools', '--version'])
    assert samtools_installed_version.startswith('samtools'), 'samtools is not installed'
    samtools_installed_version = samtools_installed_version.split('\n')[0].split(' ')[1]
    assert validate_all_conditions(samtools_installed_version, version), 'samtools version is not correct: {} incompatible with {}. Get installation instructions here: {}'.format(samtools_installed_version, version_txt, dependencies_installation_page)
    
def check_rnastructure_exists(path=None):
    version, version_txt = dependencies['rnastructure'], dependencies_txt['rnastructure']
    if path and path != None:
        cmd = [os.path.join(path, 'Fold'), '--version']
    else:
        cmd = ['Fold', '--version']
    rnastructure_installed_version = run_cmd(cmd)
    assert rnastructure_installed_version.startswith('Fold: Version '), 'RNAstructure is not found. If you are using a Mac and the "Fold" command runs in your terminal, try giving DREEM the RNAstructure absolute path through the argument "rnastructure_path" (ex: rnastructure_path="/Users/username/RNAstructure/exe").'
    assert validate_all_conditions(rnastructure_installed_version.split(' ')[2], version), 'RNAstructure version is not correct: {} incompatible with {}. Get installation instructions here: {}'.format(rnastructure_installed_version.split(' ')[2], version_txt, dependencies_installation_page)

if __name__ == '__main__':
    check_fastqc_exists()
    check_cutadapt_exists()
    check_bowtie2_exists()
    check_samtools_exists()
    check_rnastructure_exists('/Users/ymdt/src/RNAstructure/exe')
    print('All dependencies are installed correctly.')


from functools import wraps
from inspect import getmembers, Parameter, Signature
from textwrap import dedent
from typing import Any, Callable

from click import Option

from ..util import cli


# Ignore special parameters with reserved names.
reserved_params = {"self", "cls"}

# Get every option defined for the command line interface.
cli_options = dict(getmembers(cli, lambda member: isinstance(member, Option)))

# Get the default value for every parameter.
api_defs = {
    "n_procs": cli.NUM_CPUS,
}
cli_defs = {option.name: option.default for option in cli_options.values()
            if option.default is not None}
all_defs = {**cli_defs, **api_defs}

# Get the documentation for every CLI parameter.
cli_docs = {option.name: option.help for option in cli_options.values()}


def get_param_default(param: Parameter,
                      defaults: dict[str, Any],
                      exclude_defs: tuple[str, ...]) -> Parameter:
    """ Return the parameter, possibly with a new default value. """
    if param.name in exclude_defs:
        return param
    if param.name in reserved_params:
        return param
    try:
        default = defaults[param.name]
    except KeyError:
        return param
    # Return copy of parameter with new default value.
    return param.replace(default=default)


def paramdef(defaults: dict[str, Any], exclude_defs: tuple[str, ...]):
    """ Give the keyword argments of a function default values. """
    if defaults is None:
        defaults = dict()

    def decorator(func: Callable):
        # List all the parameters of the function, replacing the default
        # value of those parameters with defaults given in defaults.
        sig = Signature.from_callable(func)
        new_params = [get_param_default(param, defaults, exclude_defs)
                      for param in sig.parameters.values()]

        # Update the help text (does not affect actual default values).
        try:
            func.__signature__ = Signature(parameters=new_params)
        except ValueError as error:
            raise ValueError(f"Failed to set signature of {func.__name__} to "
                             f"({', '.join(map(str, new_params))}): {error}")

        # Update the actual default values of keyword-only arguments
        # (does not affect help text).
        default_kwargs = {param.name: param.default for param in new_params
                          if param.kind == Parameter.KEYWORD_ONLY}

        @wraps(func)
        def new_func(*args, **kwargs):
            return func(*args, **{**default_kwargs, **kwargs})

        return new_func

    return decorator


def autodef(extra_defs: dict[str, Any] | None = None,
            exclude_defs: tuple[str, ...] = ()):
    """ Call ```paramdef``` and automatically infer default values from
    the CLI and API. Extra defaults (if needed) may be given as keyword
    arguments. """
    return paramdef(all_defs if extra_defs is None
                    else {**all_defs, **extra_defs},
                    exclude_defs)


def get_param_lines(func: Callable, param_docs: dict[str, str]):
    sig = Signature.from_callable(func)
    # Add information about every parameter to the docstring.
    param_lines = list()
    for name, param in sig.parameters.items():
        if name in reserved_params:
            # Ignore reserved parameters (if any).
            continue
        if doc := param_docs.get(name):
            # Add the type annotation (if any) after the name of the
            # parameter.
            if param.annotation is param.empty:
                name_type = name
            else:
                try:
                    name_type = f"{name}: {param.annotation.__name__}"
                except AttributeError:
                    # Some types (e.g. UnionType) have no name.
                    name_type = f"{name}: {param.annotation}"
            # Add the kind of parameter and its default value (if any)
            # in brackets after the documentation of the parameter.
            doc = (f"{doc} [{param.kind.description}]"
                   if param.default is param.empty
                   else f"{doc} [{param.kind.description}, "
                        f"default: {repr(param.default)}]")
            # Add the parameter's name, type, kind, and documentation to
            # the docstring.
            param_lines.extend([f"{name_type}",
                                f"    {doc}"])
    return param_lines


def get_doc_lines(func: Callable, param_lines: list[str], return_doc: str):
    sig = Signature.from_callable(func)
    doc_lines = list()
    if func.__doc__:
        # Use the existing docstring to start the new docstring.
        doc_lines.append(dedent(func.__doc__))
    if param_lines:
        if doc_lines:
            doc_lines.append("")
        doc_lines.extend(["Parameters",
                          "----------"])
        doc_lines.extend(param_lines)
    if sig.return_annotation is not sig.empty:
        if return_doc:
            if doc_lines:
                doc_lines.append("")
            doc_lines.extend(["Return",
                              "------",
                              f"{sig.return_annotation}",
                              f"    {return_doc}"])
    return doc_lines


def paramdoc(param_docs: dict[str, str], return_doc: str):
    """
    Give a function a new docstring where each parameter gets annotated
    with the text given in the keyword arguments (```param_docs```).

    Parameters
    ----------
    param_docs: dict[str, str]
        Description of each parameter, keyed by name
    return_doc: str
        Description of the return value; will occur at end of docstring

    Return
    ------
    callable
        Function with new docstring
    """

    def decorator(func: Callable):
        param_lines = get_param_lines(func, param_docs)
        doc_lines = get_doc_lines(func, param_lines, return_doc)
        func.__doc__ = "\n".join(doc_lines)
        return func

    return decorator


def autodoc(extra_docs: dict[str, str] | None = None, return_doc: str = ""):
    """ Call ```paramdoc``` and automatically infer descriptions and
    type annotations about all parameters from the CLI and API.
    Documentation of any extra parameters may also be given. """
    return paramdoc(cli_docs if extra_docs is None
                    else {**cli_docs, **extra_docs},
                    return_doc)


def auto(*,
         extra_defs: dict[str, Any] | None = None,
         exclude_defs: tuple[str, ...] = (),
         extra_docs: dict[str, str] | None = None,
         return_doc: str = ""):
    """ Combine ```autodef``` and ```autodoc```, in that order. """

    def decorator(func: Callable):
        func = autodef(extra_defs, exclude_defs)(func)
        func = autodoc(extra_docs, return_doc)(func)
        return func

    return decorator



from ..cluster.EMclustering import EMclustering
from ..cluster.bitvector import BitVector
import numpy as np
import multiprocessing

class ClusteringAnalysis:
    """Launches the clustering algorithm many times to iterate over K_max the number of clusters and num_runs the number of runs.
    
    Parameters from args:
    ---------------------
        
    bitvector: object
        bitvector object. Conatins the processed bitvector and the count of reads per bitvector.
    
    max_clusters: int
        Number of clusters. The algorithm will iterate from 1 to max_clusters.
        
    num_runs: int
        Number of runs. The algorithm will be launched num_runs times.
        
    Key methods:
    ------------
    
    run: Run the clustering algorithm many times to iterate over K_max the number of clusters and num_runs the number of runs.
        
    """
    
    def __init__(self, bitvector:BitVector, max_clusters:int, num_runs:int, clustering_args):
        self.bitvector = bitvector
        self.K_max = max_clusters
        self.num_runs = num_runs
        self.clustering_args = clustering_args
        # TODO add all

    def run(self):
        '''Run the clustering algorithm many times to iterate over K_max the number of clusters and num_runs the number of runs.
        
        Output:
        -------

        The output is a dictionary containing the results of EMclustering for each K and each run.
        The first level of the dictionary is the number of clusters.
        The second level of the dictionary is a list containing the results of EMclustering for each run.
        The second levels of the dictionary are sorted by performance (log-likelihood).
        
        Example:
        --------
        
        results = {
            K1: [EMclustering1, EMclustering2, ...],
            K2: [EMclustering1, EMclustering2, ...],
            ...
        }
        
        The EMclustering output dictionary contains the following attributes:
            - mu: array (K x D)
                Mean of each cluster.
            - pi: array (K)
                Probability of each cluster.
            - log_likelihood: float
                Log-likelihood of the model.
        '''

        # global dT # !! For test_input !!
        em = EMclustering(self.bitvector.bv, 1, self.bitvector.read_hist, self.bitvector.base_to_keep, self.bitvector.sequence,
                                **self.clustering_args)
        results = {1: [em.run()]}
        for k in range(2, self.K_max + 1):
            em = EMclustering(self.bitvector.bv, k, self.bitvector.read_hist, self.bitvector.base_to_keep, self.bitvector.sequence,
                                **self.clustering_args)

            pool = multiprocessing.Pool(processes=self.clustering_args["max_procs"])
            results[k] = sorted(pool.starmap(em.run, [() for _ in range(self.num_runs)]), key=lambda res: res['log_likelihood'], reverse=True)
            pool.close()
            pool.join()
        
        return results
        

if __name__ == '__main__':
    pass
    # if False:

    #     import matplotlib.pyplot as plt # !! For test_input !!
    #     import time, copy # !! For test_input !!
    #     import tracemalloc # !! For test_input !!
        
    #     dT = 0 # !! For test_input !!
        
    #     # dummy class just for test_input
    #     class testBV_class:
    #         def __init__(self, bit_vector, read_hist) -> None:
    #             self.bv = bit_Vector
    #             self.read_hist = read_hist
                

    #     exp_path = "/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/"
    #     bit_Vector_total = np.load(exp_path+"bit_vector.npy")
    #     read_hist_total =  np.load(exp_path+"read_hist.npy")

    #     with open(exp_path+"data_EM_analysis_v2.txt", 'a') as f:
    #         f.write("N_reads n_cpu dT mean_memory peak_memory \n")
    #         f.close()
        
    #     for N_partial in np.geomspace(10000, 120000, 5).astype(np.int64):

    #         N_partial = min(N_partial, bit_Vector_total.shape[0])

    #         for max_procs in range(1, 11):

    #             print("Starting experiment with {} reads and {} cpus".format(N_partial, max_procs))
                
    #             tracemalloc.start()

    #             # Get part of the total reads
    #             bit_Vector = copy.copy(bit_Vector_total[:N_partial])
    #             read_hist = copy.copy(read_hist_total[:N_partial])
    #             BV_class = testBV_class(bit_vector=bit_Vector, read_hist=read_hist)
                
    #             # Run EM and record stats
    #             clustering_args = dict(
    #                 min_iter = 10,
    #                 signal_thresh = 0.5, 
    #                 info_thresh = 0.5, 
    #                 include_g_u = True, 
    #                 include_del = True, 
    #                 min_reads = 10,
    #                 convergence_cutoff = 0.5,
    #                 num_runs = 100,
    #                 max_procs = max_procs,
    #                 verbose = True
    #             )

    #             clustering = ClusteringAnalysis(BV_class, 2, 10, clustering_args)
    #             clustering.run()
                
    #             # Log experiment stats
    #             memory_stats = [ mem/1e6 for mem in tracemalloc.get_traced_memory() ]
                
    #             print("Mean and max memory [MB]:", memory_stats)
    #             print("Total time [s]", "{:.1f}s".format(dT)) # !! For test_input !!
                
    #             log_data = np.array((N_partial,
    #                                     n_cpu,
    #                                     dT,
    #                                     memory_stats[0], 
    #                                     memory_stats[1] ))
    
    #             with open(exp_path+"data_EM_analysis_v2.txt", 'a') as f:
    #                 np.savetxt(f, log_data, newline=" ")
    #                 f.write("\n")
    #                 f.close()

    #             # Stop memory tracing and clean memory
    #             tracemalloc.stop()
    #             bit_Vector = None
    #             read_hist = None
    #             time.sleep(5)
                
    #             # Comparing output with previous software -> different test
    #             # mu_reference = np.load("/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/result.npy")
                # print("Reference matched:",(mu_reference == result["mu"]).all())



import os

import pandas as pd
from click import command

from .library_samples import get_samples_info, get_library_info
from .mutation_count import generate_mut_profile_from_bit_vector
from ..util import docdef
from ..util.cli import (opt_out_dir, opt_temp_dir, opt_save_temp,
                        opt_library, opt_samples, opt_fasta,
                        opt_bv_files, opt_clustering_file,
                        opt_rnastructure_path, opt_rnastructure_use_temp,
                        opt_rnastructure_fold_args, opt_rnastructure_use_dms,
                        opt_rnastructure_dms_min_unpaired_value,
                        opt_rnastructure_dms_max_paired_value,
                        opt_rnastructure_deltag_ensemble,
                        opt_rnastructure_probability,
                        opt_verbose)
from ..util.dump import *
from ..util.files_sanity import check_library, check_samples
from ..util.rnastructure import RNAstructure
from ..vector.profile import VectorReader
from ..util.dependencies import *

params = [
    opt_fasta,
    opt_bv_files,
    opt_library,
    opt_samples,
    opt_clustering_file,
    opt_rnastructure_path,
    opt_rnastructure_use_temp,
    opt_rnastructure_fold_args,
    opt_rnastructure_use_dms,
    opt_rnastructure_dms_min_unpaired_value,
    opt_rnastructure_dms_max_paired_value,
    opt_rnastructure_deltag_ensemble,
    opt_rnastructure_probability,
    opt_verbose,
    opt_out_dir,
    opt_temp_dir,
    opt_save_temp,
]


@command("aggregate", params=params)
def cli(**kwargs):
    return run(**kwargs)


@docdef.auto()
def run( 
        fasta: str, 
        bv_files: list,
        *,
        out_dir: str,
        temp_dir: str,
        save_temp: bool,
        library: str, 
        samples: str, 
        clustering_file: str, 
        rnastructure_path: str, 
        rnastructure_use_temp: bool,
        rnastructure_fold_args: str,
        rnastructure_use_dms: str,
        rnastructure_dms_min_unpaired_value: float,
        rnastructure_dms_max_paired_value: float,
        rnastructure_deltag_ensemble: bool,
        rnastructure_probability: bool,
        verbose: bool):
    """Run the aggregate module.

    """
    
    check_rnastructure_exists(rnastructure_path)

    # Extract the arguments
    if library != '':
        library = check_library(pd.read_csv(library), fasta, out_dir) if library is not None else None
    else:
        library = None
    if samples != '':
        df_samples = check_samples(pd.read_csv(samples)) if samples != '' else None
    else:
        df_samples = None    
    
    # Make folders
    os.makedirs(out_dir, exist_ok=True)
    temp_dir = os.path.join(temp_dir, 'aggregate')
    #os.makedirs(temp_dir, exist_ok=True)

    print('Reading in bit vectors from {}...'.format(bv_files))

    reports_path = [b for b in bv_files if b.endswith('report.json')]
    
    
    for sample_path in [b for b in bv_files if not b.endswith('report.json')]:
        sample = os.path.basename(sample_path)
        for reference in os.listdir(os.path.join(sample_path)):
            for section in os.listdir(os.path.join(sample_path, reference)):
                if os.path.exists(reports_path[-1]):
                    reports_path.append(os.path.join(sample_path, reference, section, 'report.json'))
    
    all_samples = {}
    
    for report_path in reports_path:
        
        vectors = VectorReader.load(report_path)

        if vectors.sample not in all_samples:
            all_samples[vectors.sample] = {}
        if vectors.ref not in all_samples[vectors.sample]:
            all_samples[vectors.sample][vectors.ref] = {}

        all_samples[vectors.sample][vectors.ref][vectors.section] = {
            'section_start': vectors.end5,
            'section_end': vectors.end3,
            'sequence': vectors.seq.decode(),
            'pop_avg': generate_mut_profile_from_bit_vector(vectors, clustering_file=clustering_file, verbose=verbose)
        }
        
    for sample in all_samples:
        for reference in all_samples[sample]:
            # Add the library information 
            lib, section_tranlation = get_library_info(library, reference, verbose=verbose)
            all_samples[sample][reference] = {**lib, **all_samples[sample][reference]}
            
            for section in all_samples[sample][reference].copy().keys():
                if type(all_samples[sample][reference][section]) is not dict:
                    continue
                for col in ['num_aligned']:
                    all_samples[sample][reference][col] = all_samples[sample][reference][section]['pop_avg'].pop(col)
                if section in section_tranlation:
                    all_samples[sample][reference][section_tranlation[section]] = all_samples[sample][reference].pop(section)
            
    # Add the sample information
    print('Adding sample information...')
    for sample in all_samples:
        if df_samples is not None:
            all_samples[sample] = {**all_samples[sample], **get_samples_info(df_samples, sample, verbose=verbose)}
        else:
            all_samples[sample] = {**all_samples[sample], **{'sample': sample}}
    print('Done.')
    
    print('Computing confidence intervals and RNAstructure predictions...')
    
    rna = RNAstructure(rnastructure_path=rnastructure_path, temp = os.path.join(temp_dir,'rnastructure'))
    for sample, mut_profiles in all_samples.items():
        for reference in mut_profiles:
            if type(mut_profiles[reference]) is not dict:
                continue

            for section in mut_profiles[reference]:
                if type(mut_profiles[reference][section]) is not dict:
                    continue
                # Add RNAstructure predictions
                mh = rna.run(mut_profiles[reference][section]['sequence'])
                all_samples[sample][reference][section] = {**mut_profiles[reference][section], **mh}
    rna.dump_ledger()
    
    for sample, mut_profiles in all_samples.items():
        # Write the output
        out = cast_dict(mut_profiles)
        out = sort_dict(out)
        
        # Make lists in one line
        out = json.dumps(out, cls=NpEncoder, indent=2)
        backslashN = """
        """
        out = out.replace(']','[').split('[')
        out = [o.replace(backslashN+'  ','').replace(backslashN, '') if i%2 else o for i, o in enumerate(out)]
        out = out[0] + ''.join([('[',']')[i%2] + o for i, o in enumerate(out[1:])])
        
        # Write the output
        with open(os.path.join(out_dir, sample + '.json'), 'w') as f:
            f.write(out)
        print('Outputed to {}'.format(os.path.join(out_dir, sample + '.json')))
    return 1



from click import command

from ..demultiplex.demultiplex import demultiplex_run
from ..util.cli import (
    opt_barcode_length, opt_barcode_start, opt_parallel_demultiplexing, opt_clipped_demultiplexing,
    opt_mismatch_tolerence, opt_index_tolerence, opt_demulti_overwrite, opt_fasta, opt_library, opt_fastq1, opt_fastq2,opt_out_dir,opt_temp_dir)

params = [
    # Inputs
    opt_fasta,
    opt_fastq1,
    opt_fastq2,
    opt_library,
    opt_barcode_start,
    opt_barcode_length,
    opt_out_dir,
    opt_temp_dir,

    # options
    opt_parallel_demultiplexing,
    opt_clipped_demultiplexing,
    opt_mismatch_tolerence,
    opt_index_tolerence,
    opt_demulti_overwrite,


]


# Turn into DREEM command.

@command("demultiplex", params=params)
def cli(**kwargs):
    return run(**kwargs)


def run(library: str, out_dir: str, temp_dir:str, fastq1: str, fastq2: str, fasta: str, barcode_start=0,
        barcode_length=0, clipped: int = 0, index_tolerance: int = 0, parallel_demultiplexing: bool = False,
        mismatch_tolerence: int = 0, demulti_overwrite: bool = False):
    
    return demultiplex_run(library_csv=library,
                           overwrite=demulti_overwrite,
                           demulti_workspace=temp_dir,
                           report_folder=out_dir,
                           mixed_fastq1=fastq1,
                           mixed_fastq2=fastq2,
                           barcode_start=barcode_start,
                           barcode_length=barcode_length,
                           clipped=clipped,
                           index_tolerance=index_tolerance,
                           parallel=parallel_demultiplexing,
                           fasta=fasta,
                           mismatch_tolerence=mismatch_tolerence)


if __name__ == '__main__':
    pass





import os, sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

def load_dataset():
    import pandas as pd
    return pd.read_feather(os.path.join(os.path.dirname(__file__), 'resources/my_dataset.feather'))




from setuptools import setup, find_packages


requirements = []
with open('dreem/requirements.txt', 'r') as fh:
    for line in fh:
        requirements.append(line.strip())

# PYTHON_VERSION = (3,10)

# if sys.version_info < PYTHON_VERSION:
#    sys.exit(f"Python >= {PYTHON_VERSION[0]}.{PYTHON_VERSION[1]} required.")

readme = open('README.md').read()

setup(
    name="dreem",
    version='0.1.8',
    description="DREEM solves RNA structure ensembles using chemical probing data",
    long_description=readme,
    author="Silvi Rouskin Lab",
    author_email="silvi@hms.harvard.edu",
    url="https://github.com/rouskinlab/dreem",
    packages=find_packages(),
    package_dir={'dreem': 'dreem'},
    include_package_data=True,
    package_data={
        "dreem": ["test-data/vector-test-data/vectorize-read-test-data.csv"]
    },
    install_requires=requirements,
    entry_points={'console_scripts': ['dreem = dreem.main : cli']},
    python_requires=">=3.10",
)



from logging import getLogger
import pathlib

from click import command

from .fq2bam import get_bam_files
from .fqs import FastqUnit
from ..util import docdef
from ..util.cli import (opt_fasta,
                        opt_fastqs, opt_fastqi, opt_fastq1, opt_fastq2,
                        opt_fastqs_dir, opt_fastqi_dir, opt_fastq12_dir,
                        opt_phred_enc,
                        opt_out_dir, opt_temp_dir,
                        opt_rerun, opt_save_temp,
                        opt_parallel, opt_max_procs,
                        opt_fastqc, opt_qc_extract,
                        opt_cutadapt,
                        opt_cut_a1, opt_cut_g1, opt_cut_a2, opt_cut_g2,
                        opt_cut_o, opt_cut_e, opt_cut_q1, opt_cut_q2,
                        opt_cut_m, opt_cut_indels,
                        opt_cut_discard_trimmed, opt_cut_discard_untrimmed,
                        opt_cut_nextseq,
                        opt_bt2_local, opt_bt2_unal,
                        opt_bt2_discordant, opt_bt2_mixed,
                        opt_bt2_dovetail, opt_bt2_contain,
                        opt_bt2_i, opt_bt2_x, opt_bt2_score_min,
                        opt_bt2_s, opt_bt2_l, opt_bt2_d, opt_bt2_r,
                        opt_bt2_gbar, opt_bt2_dpad, opt_bt2_orient)

from ..util.parallel import lock_output


logger = getLogger(__name__)

from ..util.dependencies import check_bowtie2_exists, check_cutadapt_exists, check_fastqc_exists, check_samtools_exists


# Parameters for command line interface
params = [
    # Inputs
    opt_fasta,
    opt_fastqs,
    opt_fastqi,
    opt_fastq1,
    opt_fastq2,
    opt_fastqs_dir,
    opt_fastqi_dir,
    opt_fastq12_dir,
    opt_phred_enc,
    # Outputs
    opt_out_dir,
    opt_temp_dir,
    opt_rerun,
    opt_save_temp,
    # Parallelization
    opt_parallel,
    opt_max_procs,
    # FASTQC
    opt_fastqc,
    opt_qc_extract,
    # Cutadapt
    opt_cutadapt,
    opt_cut_a1,
    opt_cut_g1,
    opt_cut_a2,
    opt_cut_g2,
    opt_cut_o,
    opt_cut_e,
    opt_cut_q1,
    opt_cut_q2,
    opt_cut_m,
    opt_cut_indels,
    opt_cut_discard_trimmed,
    opt_cut_discard_untrimmed,
    opt_cut_nextseq,
    # Bowtie2
    opt_bt2_local,
    opt_bt2_discordant,
    opt_bt2_mixed,
    opt_bt2_dovetail,
    opt_bt2_contain,
    opt_bt2_unal,
    opt_bt2_i,
    opt_bt2_x,
    opt_bt2_score_min,
    opt_bt2_s,
    opt_bt2_l,
    opt_bt2_gbar,
    opt_bt2_d,
    opt_bt2_r,
    opt_bt2_dpad,
    opt_bt2_orient,
]


@command("align", params=params)
def cli(**kwargs):
    return run(**kwargs)


@lock_output
@docdef.auto()
def run(*,
        # Inputs
        fasta: str,
        fastqs: tuple[str],
        fastqi: tuple[str],
        fastq1: tuple[str],
        fastq2: tuple[str],
        fastqs_dir: tuple[str],
        fastqi_dir: tuple[str],
        fastq12_dir: tuple[str],
        phred_enc: int,
        # Outputs
        out_dir: str,
        temp_dir: str,
        save_temp: bool,
        rerun: bool,
        # Parallelization
        max_procs: int,
        parallel: bool,
        # FASTQC
        fastqc: bool,
        qc_extract: bool,
        # Cutadapt
        cut: bool,
        cut_q1: int,
        cut_q2: int,
        cut_g1: tuple[str],
        cut_a1: tuple[str],
        cut_g2: tuple[str],
        cut_a2: tuple[str],
        cut_o: int,
        cut_e: float,
        cut_indels: bool,
        cut_nextseq: bool,
        cut_discard_trimmed: bool,
        cut_discard_untrimmed: bool,
        cut_m: int,
        # Bowtie2
        bt2_local: bool,
        bt2_discordant: bool,
        bt2_mixed: bool,
        bt2_dovetail: bool,
        bt2_contain: bool,
        bt2_unal: bool,
        bt2_score_min: str,
        bt2_i: int,
        bt2_x: int,
        bt2_gbar: int,
        bt2_l: int,
        bt2_s: str,
        bt2_d: int,
        bt2_r: int,
        bt2_dpad: int,
        bt2_orient: str) -> tuple[str, ...]:
    """
    Run the alignment module.

    Align the reads to the set of reference sequences and output one BAM file
    for each sample aligned to each reference in the directory 'output'.
    Temporary intermediary files are written in the directory 'temp' and then
    deleted after they are no longer needed.
    """

    check_bowtie2_exists()
    check_cutadapt_exists()
    check_fastqc_exists()
    check_samtools_exists()
    
    if not fasta:
        logger.critical("No FASTA file was given to alignment")
        return ()

    # FASTQ files of read sequences may come from up to seven different
    # sources (i.e. each argument beginning with "fq_unit"). This step
    # collects all of them into one list (fq_units) and also bundles
    # together pairs of FASTQ files containing mate 1 and mate 2 reads.
    fq_units = list(FastqUnit.from_strs(fastqs=fastqs,
                                        fastqi=fastqi,
                                        fastq1=fastq1,
                                        fastq2=fastq2,
                                        fastqs_dir=fastqs_dir,
                                        fastqi_dir=fastqi_dir,
                                        fastq12_dir=fastq12_dir,
                                        phred_enc=phred_enc))

    # Generate and return a BAM file for every FASTQ-reference pair.
    return get_bam_files(fq_units=fq_units,
                         fasta=pathlib.Path(fasta),
                         out_dir=pathlib.Path(out_dir),
                         temp_dir=pathlib.Path(temp_dir),
                         save_temp=save_temp,
                         rerun=rerun,
                         max_procs=max_procs,
                         parallel=parallel,
                         fastqc=fastqc,
                         qc_extract=qc_extract,
                         cut=cut,
                         cut_q1=cut_q1,
                         cut_q2=cut_q2,
                         cut_g1=cut_g1,
                         cut_a1=cut_a1,
                         cut_g2=cut_g2,
                         cut_a2=cut_a2,
                         cut_o=cut_o,
                         cut_e=cut_e,
                         cut_indels=cut_indels,
                         cut_nextseq=cut_nextseq,
                         cut_discard_trimmed=cut_discard_trimmed,
                         cut_discard_untrimmed=cut_discard_untrimmed,
                         cut_m=cut_m,
                         bt2_local=bt2_local,
                         bt2_discordant=bt2_discordant,
                         bt2_mixed=bt2_mixed,
                         bt2_dovetail=bt2_dovetail,
                         bt2_contain=bt2_contain,
                         bt2_unal=bt2_unal,
                         bt2_score_min=bt2_score_min,
                         bt2_i=bt2_i,
                         bt2_x=bt2_x,
                         bt2_gbar=bt2_gbar,
                         bt2_l=bt2_l,
                         bt2_s=bt2_s,
                         bt2_d=bt2_d,
                         bt2_r=bt2_r,
                         bt2_dpad=bt2_dpad,
                         bt2_orient=bt2_orient)



from logging import getLogger
import pathlib
from typing import Iterable

from click import command
import pandas as pd

from ..util import docdef, path
from ..util.cli import (opt_fasta, opt_bamf, opt_bamd,
                        opt_library, opt_autosect,
                        opt_coords, opt_primers, opt_primer_gap,
                        opt_out_dir, opt_temp_dir,
                        opt_phred_enc, opt_min_phred,
                        opt_strict_pairs, opt_ambid, opt_batch_size,
                        opt_parallel, opt_max_procs,
                        opt_rerun, opt_save_temp)
from ..util.parallel import lock_output
from ..util.seq import DNA
from ..vector.profile import generate_profiles, get_writers


logger = getLogger(__name__)


def add_coords_from_library(library_path: str,
                            initial_coords: tuple[tuple[str, int, int]]):
    library_coords = list()
    try:
        library = pd.read_csv(library_path)
        for ref, end5, end3 in zip(library["reference"],
                                   library["section_start"],
                                   library["section_end"], strict=True):
            try:
                coord = (str(pathlib.Path(ref).with_suffix("")),
                         int(end5),
                         int(end3))
            except ValueError as error:
                logger.error(f"Failed to add coordinates {ref, end5, end3} "
                              f"with the following error: {error}")
            else:
                if coord in initial_coords or coord in library_coords:
                    logger.warning(f"Skipping duplicate coordinates: {coord}")
                else:
                    library_coords.append(coord)
    except (FileNotFoundError, KeyError, ValueError) as error:
        logger.critical(f"Failed to add coordinates from {library_path} "
                      f"with the following error: {error}")
    return initial_coords + tuple(library_coords)


def encode_primers(primers: Iterable[tuple[str, str, str]]):
    for ref, fwd, rev in primers:
        try:
            yield ref, DNA(fwd.encode()), DNA(rev.encode())
        except ValueError as error:
            logger.error(f"Failed to add primer pair {ref, fwd, rev} "
                          f"with the following error: {error}")


def list_bam_paths(bamf: tuple[str, ...], bamd: tuple[str, ...]):
    bam_files: set[pathlib.Path] = set()

    def add_bam_file(file: pathlib.Path):
        if not file.is_file():
            logger.critical(f"Skipping non-existant BAM file: {file}")
            return
        if file.suffix != path.BAM_EXT:
            logger.critical(f"Skipping non-BAM-formatted file: {file}")
            return
        if file in bam_files:
            logger.warning(f"Skipping duplicate BAM file: {file}")
            return
        bam_files.add(file)

    for bam_file in bamf:
        add_bam_file(pathlib.Path(bam_file))
    for bam_dir in bamd:
        dpath = pathlib.Path(bam_dir)
        if not dpath.is_dir():
            logger.critical(f"Skipping non-existant BAM directory: {dpath}")
            continue
        for bamd_file in dpath.iterdir():
            add_bam_file(bamd_file)

    return list(bam_files)


# Parameters for command line interface
params = [
    # Input files
    opt_fasta,
    opt_bamf,
    opt_bamd,
    # Sections
    opt_library,
    opt_autosect,
    opt_coords,
    opt_primers,
    opt_primer_gap,
    # Output directories
    opt_out_dir,
    opt_temp_dir,
    # SAM options
    opt_phred_enc,
    opt_min_phred,
    # Vectoring options
    opt_ambid,
    opt_strict_pairs,
    opt_batch_size,
    # Parallelization
    opt_max_procs,
    opt_parallel,
    # File generation
    opt_rerun,
    opt_save_temp,
]


@command("vector", params=params)
def cli(**kwargs):
    """ Hook the command line interface to the ```run``` function. """
    return run(**kwargs)


@lock_output
@docdef.auto()
def run(fasta: str,
        *,
        bamf: tuple[str],
        bamd: tuple[str],
        library: str,
        autosect: bool,
        coords: tuple[tuple[str, int, int], ...],
        primers: tuple[tuple[str, str, str], ...],
        primer_gap: int,
        out_dir: str,
        temp_dir: str,
        phred_enc: int,
        min_phred: int,
        ambid: bool,
        strict_pairs: bool,
        batch_size: float,
        max_procs: int,
        parallel: bool,
        rerun: bool,
        save_temp: bool):
    """ Run the vectoring step. Generate a vector encoding mutations for
    each read (or read pair, if paired-end). """

    if not fasta:
        logger.critical("No FASTA file was given to vectoring")
        return ()

    # If a library file is given, add coordinates from the file.
    if library:
        coords_lib = add_coords_from_library(library_path=library,
                                             initial_coords=coords)
    else:
        coords_lib = coords

    # List all BAM files among the given files and directories.
    bams = list_bam_paths(bamf, bamd)

    # Create an object to write mutation vectors for each BAM file.
    writers, temp_bams = get_writers(pathlib.Path(fasta), bams,
                                     temp_dir=pathlib.Path(temp_dir),
                                     n_procs=max_procs,
                                     coords=coords_lib,
                                     primers=tuple(encode_primers(primers)),
                                     primer_gap=primer_gap,
                                     autosect=autosect)

    try:
        # Compute and write mutation vectors for each BAM file.
        profiles = generate_profiles(writers=writers,
                                     out_dir=pathlib.Path(out_dir),
                                     temp_dir=pathlib.Path(temp_dir),
                                     phred_enc=phred_enc,
                                     min_phred=min_phred,
                                     ambid=ambid,
                                     strict_pairs=strict_pairs,
                                     batch_size=batch_size,
                                     max_procs=max_procs,
                                     parallel=parallel,
                                     rerun=rerun,
                                     save_temp=save_temp)
        return profiles
    finally:
        if not save_temp:
            # Always delete the temporary BAM and index files, whether
            # vectoring completed normally or raised an error.
            for temp_bam in temp_bams:
                temp_bam.unlink(missing_ok=True)



import os, sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
import os
import pytest
import dreem

test_files = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'test_files')
test_files_sample = os.path.join(test_files, 'my_sample_to_demultiplex')
demultiplexed_fastq_dir = os.path.join(test_files_sample,'my_sample_to_demultiplex')
library = os.path.join(test_files_sample, 'library.csv')
fasta =  os.path.join(test_files_sample, 'my_sample_to_demultiplex.fasta')
fastq1 = os.path.join(test_files_sample, 'my_sample_to_demultiplex_R1.fastq')
fastq2 = os.path.join(test_files_sample, 'my_sample_to_demultiplex_R2.fastq')
out_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'test_output','demultiplexing')
temp_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'test_temp','demultiplexing')

def test_demultiplexing_python():
    dreem.demultiplex.run(
        library=library,
        fastq1=(fastq1,),
        fastq2=(fastq2,),
        clipped=2,
        index_tolerance=10,
        mismatch_tolerence=2,
        out_dir=os.path.join(out_dir, 'python'),
        temp_dir=os.path.join(out_dir, 'python'),
        fasta=fasta,
    )

def test_demultiplexing_cli():
    os.system(f"dreem demultiplex --library {library} --fasta {fasta} --fastq1 {fastq1} --fastq2 {fastq1} --out-dir {os.path.join(out_dir, 'cli')} --temp-dir {os.path.join(temp_dir, 'cli')} --index-tolerance {10} --mismatch-tolerence {2} ")
    
@pytest.mark.parametrize('sample', ['cli', 'python'])
def test_results(sample):
    directory = os.path.join(out_dir, sample, 'my_sample_to_demultiplex')
    for file in os.listdir(directory):
        lines = open(os.path.join(directory, file)).readlines()
        assert len(lines) == 4*9 or len(lines) == 4*9+1, 'File {} has {} lines, expected {}'.format(file, len(lines), 4*10) # 4 lines per read, 10 reads per sample, the 10th read is garbage
    
if __name__ == '__main__':
    test_demultiplexing_python()
    test_demultiplexing_cli()


from __future__ import annotations

from collections import defaultdict, namedtuple
from datetime import datetime, timedelta
from functools import cache, cached_property, partial
from inspect import isclass
from itertools import chain, starmap as itsmap
import json
from logging import getLogger
from multiprocessing import Pool
from pathlib import Path
import re
from sys import byteorder
from time import sleep
from typing import Any, Sequence

import numpy as np
import pandas as pd

from ..align.xams import index_bam
from ..util import path
from ..util.parallel import get_num_parallel
from ..util.seq import (BLANK, MATCH, DELET, INS_5, INS_3,
                        SUB_A, SUB_C, SUB_G, SUB_T, EVERY,
                        BASES, DNA, parse_fasta)
from ..util.util import digest_file
from ..vector.samread import SamReader
from ..vector.vector import vectorize_line, vectorize_pair, VectorError

logger = getLogger(__name__)

SectionTuple = namedtuple("PrimerTuple", ["pos5", "pos3"])


def mib_to_bytes(batch_size: float):
    """
    Return the number of bytes per batch of a given size in mebibytes.

    Parameters
    ----------
    batch_size: float
        Size of the batch in mebibytes (1 MiB = 2^20 bytes)

    Return
    ------
    int
        Number of bytes per batch, to the nearest integer
    """
    return round(batch_size * 1048576)  # 1048576 = 2^20


class Section(object):
    """
    Represent a section of a reference sequence between two coordinates.

    Attributes
    ----------
    ref: str
        Name of the reference sequence
    end5: int (1 ≤ end5 ≤ end3)
        Coordinate of the reference sequence at which the section's
        5' end is located (1-indexed)
    end3: int (end5 ≤ end3 ≤ len(ref_seq))
        Coordinate of the reference sequence at which the section's
        3' end is located (1-indexed; end3 itself is included)
    seq: DNA
        Sequence of the section between end5 and end3 (inclusive)
    isfullref: bool
        Whether the section sequence is the full reference sequence

    Examples
    --------
    >>> seq = DNA(b"CATCTGGA")
    >>> name = "example"
    >>> sect = Section(ref_seq=seq, ref=name, end5=1, end3=8)
    >>> assert sect.seq == seq
    >>> sect = Section(ref_seq=seq, ref=name, end5=1, end3=-1)
    >>> assert sect.seq == seq
    >>> sect = Section(ref_seq=seq, ref=name, end5=-8, end3=8)
    >>> assert sect.seq == seq
    >>> sect = Section(ref_seq=seq, ref=name, end5=3, end3=7)
    >>> assert sect.seq == DNA(b"TCTGG")
    >>> sect = Section(ref_seq=seq, ref=name, end5=-5, end3=-3)
    >>> assert sect.seq == DNA(b"CTG")
    >>> try:
    ...     sect = Section(ref_seq=seq, ref=name, end5=-9, end3=5)
    ...     assert False, "Failed to catch end5 < -len(ref_seq)"
    ... except ValueError:
    ...     pass
    >>> try:
    ...     sect = Section(ref_seq=seq, ref=name, end5=6, end3=5)
    ...     assert False, "Failed to catch end3 < end5"
    ... except ValueError:
    ...     pass
    >>> try:
    ...     sect = Section(ref_seq=seq, ref=name, end5=1, end3=9)
    ...     assert False, "Failed to catch end3 > len(ref_seq)"
    ... except ValueError:
    ...     pass
    """

    def __init__(self, /, *,
                 ref: str,
                 end5: int,
                 end3: int,
                 ref_seq: DNA | None = None,
                 sect_seq: DNA | None = None,
                 isfullref: bool = False):
        """
        Parameters
        ----------
        ref: str
            Name of the reference sequence
        end5: int (-len(ref_seq) ≤ end5 ≤ len(ref_seq); end5 ≠ 0)
            Coordinate of the reference sequence at which the 5' end of
            the section is located. If positive, number the coordinates
            of the reference sequence 1, 2, ... starting at the 5' end
            (i.e. 1-based indexing). If negative, number the coordinates
            of the reference sequence -1, -2, ... starting at the 3' end
            (i.e. 1-based indexing from the other side), then convert to
            the corresponding (positive) 1-based index from the 5' end
        end3: int (-len(ref_seq) ≤ end3 ≤ len(ref_seq); end3 ≠ 0)
            Coordinate of the reference sequence at which the section's
            3' end is located. Follows the same coordinate numbering
            convention as end5
        ref_seq: DNA | None
            Sequence of the entire reference (must provide either
            ```ref_seq``` or ```sect_seq```, but not both)
        sect_seq: DNA | None
            Sequence of the section only (must provide either
            ```ref_seq``` or ```sect_seq```, but not both)
        """
        self.ref = ref
        if ref_seq and sect_seq:
            raise ValueError("Cannot give both ref_seq and sect_seq")
        if ref_seq:
            if end5 < 0:
                # Compute the corresponding positive coordinate.
                end5 += len(ref_seq) + 1
            if end3 < 0:
                # Compute the corresponding positive coordinate.
                end3 += len(ref_seq) + 1
            self.end5 = end5
            self.end3 = end3
            if not 1 <= end5 <= end3 <= len(ref_seq):
                raise ValueError("Must have 1 ≤ end5 ≤ end3 ≤ len(ref_seq), "
                                 f"but got end5 = {end5}, end3 = {end3}, and "
                                 f"len(ref_seq) = {len(ref_seq)}")
            self.seq = ref_seq[end5 - 1: end3]
            self.isfullref = self.seq == ref_seq
        elif sect_seq:
            self.end5 = end5
            self.end3 = end3
            if not 1 <= end5 <= end3:
                raise ValueError("Must have 1 ≤ end5 ≤ end3 ≤ len(ref_seq), "
                                 f"but got end5 = {end5}, end3 = {end3}")
            if not self.length == len(sect_seq):
                raise ValueError(f"Calculated length of {self.length} from "
                                 f"end5 = {end5} and end3 = {end3}, but got "
                                 f"sect_seq of length {len(sect_seq)}")
            self.seq = sect_seq
            self.isfullref = isfullref
        else:
            raise ValueError("Must give either ref_seq or sect_seq")

    @property
    def path_fields(self):
        return {"ref": self.ref, "end5": self.end5, "end3": self.end3}

    @property
    def report_fields(self):
        return {"ref": self.ref, "end5": self.end5, "end3": self.end3,
                "isfullref": self.isfullref}

    @property
    def length(self):
        """ Return the length of the section of interest. """
        return self.end3 - self.end5 + 1

    @cached_property
    def positions(self):
        """ Return all positions in the section of interest. """
        return np.arange(self.end5, self.end3 + 1)

    def subseq(self, positions: Sequence[int] | None):
        """ Return a subset of the sequence at the given positions, or
        the entire sequence if positions is None. """
        if positions is None:
            # Return the entire sequence if no positions are selected.
            return self.seq
        n_pos = len(positions)
        if n_pos == 0:
            raise ValueError("Positions is an empty sequence")
        pos5, pos3 = positions[0], positions[-1]
        if n_pos != pos3 - pos5 + 1:
            raise ValueError(
                "Positions must be a sequence of monotonically increasing "
                f"consecutive integers, but got {positions}")
        if pos5 < self.end5:
            raise ValueError(f"5' end ({pos5}) out of bounds for {self}")
        if pos3 > self.end3:
            raise ValueError(f"3' end ({pos3}) out of bounds for {self}")
        return self.seq[pos5 - self.end5: pos3 - self.end5 + 1]

    @property
    def ref_coords(self):
        """ Return the name of the reference and the 5' and 3' positions
        of the section of interest; for hashing and equality test_input. """
        return self.ref, self.end5, self.end3

    @staticmethod
    def seq_pos_to_cols(seq: bytes, positions: Sequence[int]):
        """ Convert sequence and positions to column names. Each column
        name is a string of the base followed by the position. """
        # Use chr(base) instead of base.decode() because base is an int.
        return [f"{chr(base)}{pos}" for base, pos in zip(seq, positions,
                                                         strict=True)]

    @staticmethod
    def cols_to_seq_pos(columns: list[str]):
        """ Convert column names to sequence and positions. Each column
        name is a string of the base followed by the position. """
        # Regex pattern "^([ACGT])([0-9]+)$" finds the base and position
        # in each column name.
        pattern = re.compile(f"^([{BASES.decode()}])([0-9]+)$")
        # Match each column name using the pattern.
        matches = list(map(pattern.match, columns))
        try:
            # Obtain the two groups (base and position) from each match
            # and unzip them into two tuples.
            bases, pos_strs = zip(*map(re.Match.groups, matches))
        except TypeError:
            # TypeError is raised if any match is None, which happens if
            # a column fails to match the pattern.
            invalid_cols = [col for col, match in zip(columns, matches,
                                                      strict=True)
                            if match is None]
            raise ValueError(f"Invalid columns: {invalid_cols}")
        # Join the tuple of bases into a DNA sequence.
        seq = DNA("".join(bases).encode())
        # Cast the tuple of strings of positions into an integer array.
        positions = np.array(list(map(int, pos_strs)))
        return seq, positions

    @cached_property
    def columns(self):
        """ Return the column names of the section. """
        return self.seq_pos_to_cols(self.seq, self.positions)

    @property
    def tag(self):
        """ Return a hashable identifier for the section. """
        return self.ref, self.end5, self.end3

    @property
    def section(self):
        return f"{self.end5}-{self.end3}"

    def __str__(self):
        return f"{self.ref}:{self.section}"


class SectionFinder(Section):
    """
    The 5' and 3' ends of a section can be given explicitly as integers, but if
    the sample is of an amplicon (i.e. generated by RT-PCR using site-specific
    primers), then it is often more convenient to enter the sequences of the
    PCR primers and have the software determine the coordinates. SectionFinder
    accepts 5' and 3' coordinates given as integers or primers, validates them,
    and stores the coordinates as integers. This process works as follows:

    end5
    - If given as a parameter, this value is used.
    - Else if fwd is given, first is based on the location of its 3' end.
      - Else first is set to 1.
        - last
          - If last is given as an argument, last is set to this value.
          - Else if rev is given, last is based on the location of the 5' end
            of its reverse complement.
          - Else last is set to the length of ref_seq.

    """

    def __init__(self, /, *, ref_seq: DNA | None, ref: str, primer_gap: int,
                 end5: int | None = None, end3: int | None = None,
                 fwd: DNA | None = None, rev: DNA | None = None):
        """
        Parameters
        ----------
        ref_seq: see superclass
        ref: see superclass
        primer_gap: int
            (For coordinates specified by fwd/rev only) Number of
            positions 3' of the forward primer and 5' of the reverse
            primer to exclude from the section. Coordinates within 1 - 2
            nucleotides of each primer may contain DMS reactivity
            artifacts. If primer_gap = 0, then end5 and end3 are set,
            respectively, to the coordinates immediately adjacent to
            (i.e. 1 nucleotide 3' and 5' of) the 3' end of the forward
            and reverse primers.
        end5: int | None (default: None)
            If given, behaves as in the superclass; otherwise, ignored.
        end3: int | None (default: None)
            If given, behaves as in the superclass; otherwise, ignored.
        fwd: DNA | None = None (default: None)
            (For amplicons only) Sequence of the forward PCR primer
            that was used to generate the amplicon
        rev: DNA | None = None (default: None)
            (For amplicons only) Sequence of the reverse PCR primer
            that was used to generate the amplicon (the actual sequence,
            not its reverse complement)
        """
        if primer_gap < 0:
            logger.warning("Primer gap must be ≥ 0: setting to 0")
            primer_gap = 0
        if ref_seq is None:
            raise ValueError(f"No sequence for reference named '{ref}'. Check "
                             "that you gave the right reference sequence file "
                             "and spelled the name of the reference correctly.")
        offset = primer_gap + 1
        if end5 is None:
            # If first is to be determined from the fwd primer sequence,
            # the primer is aligned to the reference, and first is set to
            # the position (primer_gap + 1) downstream of its 3' coordinate.
            end5 = (1 if fwd is None
                    else self.locate(ref_seq, fwd).pos3 + offset)
        if end3 is None:
            # If last is to be determined from the rev primer sequence,
            # the reverse complement of the primer is aligned to the reference,
            # and last is set to the position (primer_gap + 1) upstream of its
            # 5' coordinate.
            end3 = (len(ref_seq) if rev is None
                    else self.locate(ref_seq, rev.rc).pos5 - offset)
        super().__init__(ref_seq=ref_seq, ref=ref, end5=end5, end3=end3)

    @staticmethod
    def locate(ref_seq: DNA, primer: DNA) -> SectionTuple:
        """
        Return the 5' and 3' positions (1-indexed) of a primer within a
        reference sequence. The primer must occur exactly once in the
        reference, otherwise an error is raised.

        Parameters
        ----------
        ref_seq: DNA
            Sequence of the entire reference (not just the section of interest)
        primer: DNA
            Sequence of the forward PCR primer or of the reverse complement of
            the reverse PCR primer
        
        Returns
        -------
        SectionTuple
            Named tuple of the first and last positions that the primer occupies
            in the reference sequence. Positions are 1-indexed and include the
            first and last coordinates.
        """
        matches = list(re.finditer(primer, ref_seq))
        if not matches:
            raise ValueError(f"Primer '{primer}' is not in ref '{ref_seq}'")
        if len(matches) > 1:
            raise ValueError(f"Primer '{primer}' occurs {len(matches)} times "
                             f"in ref '{ref_seq}'")
        # Add 1 to convert from 0-indexed (re.finditer) to 1-indexed (DREEM).
        pos5 = matches[0].start() + 1
        # No change is needed to convert from exclusive 0-indexed (re.finditer)
        # to inclusive 1-indexed (DREEM).
        pos3 = matches[0].end()
        return SectionTuple(pos5, pos3)


class MutationalProfile(Section):
    """
    Represent all reads from one sample that overlap a section of interest in a
    reference sequence.

    Fields
    ------
    sample: str
        Name of the sample
    """

    def __init__(self, /, *, sample: str, **kwargs):
        super().__init__(**kwargs)
        self.sample = sample

    @property
    def path_fields(self):
        return {"module": path.Module.VECTOR.value,
                "sample": self.sample,
                **super().path_fields}

    @property
    def report_fields(self):
        return {"sample": self.sample,
                **super().report_fields}

    @property
    def tag(self):
        return tuple([self.sample, *super().tag])

    def get_batch_dir(self, out_dir: str):
        """ Directory in which all batches of mutation vectors are
        written. """
        return path.SectionOutDirPath(top=out_dir, **self.path_fields)

    def get_mv_batch_path(self, out_dir: Path, batch: int):
        """ File in which one batch of mutation vectors is written. """
        return path.MutVectorBatchFilePath(top=str(out_dir),
                                           **self.path_fields,
                                           batch=batch,
                                           ext=".orc")

    def __str__(self):
        return f"{self.sample}@{super().__str__()}"


class VectorWriter(MutationalProfile):
    """
    Compute mutation vectors for all reads from one sample mapping to one
    section of one reference sequence.
    """

    def __init__(self, /, bam_file: Path, **kwargs):
        bam_path = path.OneRefAlignmentInFilePath.parse(bam_file)
        super().__init__(sample=bam_path.sample,
                         ref=bam_path.ref,
                         **kwargs)
        self.bam_path = bam_path
        self.byteseq = bytes(self.seq)

    def get_report_path(self, out_dir: Path):
        return path.MutVectorReportFilePath(top=str(out_dir),
                                            **self.path_fields,
                                            ext=".json")

    def outputs_valid(self, /, out_dir: Path):
        """ Return whether the report file exists and, if so, whether
        all batch files of mutation vectors listed in the report exist
        and match the checksums recorded in the report. """
        try:
            VectorReport.load(self.get_report_path(out_dir).path)
        except (FileNotFoundError, ValueError):
            return False
        else:
            return True

    def _write_report(self, /, *, out_dir: Path, **kwargs):
        logger.info(f"Began writing report of {self}")
        report = VectorReport(seq=self.seq, **self.report_fields, **kwargs)
        report_path = report.save(out_dir)
        logger.info(f"Ended writing report of {self} to {report_path}")
        return report_path

    def _write_batch(self, /,
                     out_dir: Path,
                     batch_num: int,
                     read_names: list[str],
                     vectors: tuple[bytearray, ...],
                     n_passed: int):
        """ Write a batch of mutation vectors to an ORC file. """
        # Process the mutation vectors into a 2D NumPy array.
        logger.info(f"Began writing batch {batch_num} of {self}")
        array = np.frombuffer(b"".join(vectors), dtype=np.byte)
        array.shape = (n_passed, self.length)
        # Data must be converted to pd.DataFrame for PyArrow to write.
        # Set copy=False to prevent copying the mutation vectors.
        mut_frame = pd.DataFrame(data=array,
                                 index=read_names,
                                 columns=self.columns,
                                 copy=False)
        mv_file = self.get_mv_batch_path(out_dir, batch_num)
        mv_file.path.parent.mkdir(parents=True, exist_ok=True)
        mut_frame.to_orc(mv_file.path, index=True, engine="pyarrow")
        logger.info(f"Ended writing batch {batch_num} of {self} to {mv_file}")
        return mv_file

    def _vectorize_record(self, /,
                          read_name: bytes, line1: bytes, line2: bytes, *,
                          min_qual: int, ambid: bool):
        """ Compute the mutation vector of a record in a SAM file. """
        try:
            # Initialize a blank mutation vector of (self.length) bytes.
            muts = bytearray(self.length)
            # Fill the mutation vector with data from the SAM line(s).
            if line2:
                # Using seq instead of byteseq crashes vectoring.
                vectorize_pair(line1, line2, muts, self.byteseq, self.length,
                               self.end5, self.ref, min_qual, ambid)
            else:
                # Using seq instead of byteseq crashes vectoring.
                vectorize_line(line1, muts, self.byteseq, self.length,
                               self.end5, self.ref, min_qual, ambid)
            if not any(muts):
                raise VectorError(f"{read_name} gave a blank mutation vector")
        except VectorError as error:
            logger.error(
                f"Read '{read_name.decode()}' failed to vectorize: {error}")
            read_name = b""
            muts = bytearray()
        return read_name, muts

    def _vectorize_records(self, /, *,
                           reader: SamReader,
                           start: int,
                           stop: int,
                           strict_pairs: bool,
                           phred_enc: int,
                           min_phred: int,
                           ambid: bool):
        """ Generate a batch of mutation vectors, and return them along
        with the name of every vector. """
        with reader as reading:
            # Use the SAM reader to generate the mutation vectors.
            # Collect them as a single, 1-dimensional bytes object.
            vectorize_record = partial(self._vectorize_record,
                                       min_qual=get_min_qual(min_phred,
                                                             phred_enc),
                                       ambid=ambid)
            iter_records = reading.iter_records(start, stop, strict_pairs)
            read_names, muts = zip(*itsmap(vectorize_record, iter_records))
            # For every read for which creating a mutation vector
            # failed, an empty string was returned as the read name
            # and an empty bytearray as the mutation vector. The
            # empty read names must be filtered out, while the empty
            # mutation vectors will not cause problems because,
            # being of length zero, they will effectively disappear
            # when all the vectors are concatenated into a 1D array.
            read_names = list(filter(None, read_names))
        return read_names, muts

    def _vectorize_batch(self, /,
                         batch_num: int,
                         reader: SamReader,
                         start: int,
                         stop: int, *,
                         out_dir: Path,
                         **kwargs):
        """ Compute mutation vectors for every SAM record in one batch,
        write the vectors to a batch file, and return its MD5 checksum
        and the number of vectors. """
        logger.debug(f"Began vectorizing batch {batch_num} of {self} "
                     f"({start} - {stop})")
        # Compute the read names and mutation vectors.
        read_names, muts = self._vectorize_records(reader=reader,
                                                   start=start,
                                                   stop=stop,
                                                   **kwargs)
        # Compute the number of reads that passed and failed.
        n_reads = len(muts)
        n_pass = len(read_names)
        n_fail = n_reads - n_pass
        if n_pass == 0:
            logger.critical(f"Failed to assemble batch {batch_num} of {self}")
            return n_pass, n_fail, ""
        # Write the names and vectors to a file.
        b_file = self._write_batch(out_dir, batch_num, read_names, muts, n_pass)
        if b_file is None:
            logger.critical(f"Failed to assemble batch {batch_num} of {self}")
            return n_pass, n_fail, ""
        # Compute the MD5 checksum of the file.
        checksum = digest_file(b_file.path)
        logger.debug(f"Ended vectorizing batch {batch_num} of {self} "
                     f"({start} - {stop}): {n_pass} pass, {n_fail} fail")
        return n_pass, n_fail, checksum

    def _vectorize_bam(self, /, *,
                       out_dir: Path,
                       temp_dir: Path,
                       batch_size: int,
                       n_procs: int,
                       save_temp: bool,
                       **kwargs):
        """ Compute a mutation vector for every record in a BAM file,
        split among one or more batches. For each batch, write the
        vectors to one batch file, and compute its checksum. """
        # Open the primary SAM file reader to write the subset of SAM
        # records to a temporary SAM file and determine the number and
        # start/stop indexes of each batch of records in the file.
        # The SAM file will remain open until exiting the with block.
        logger.info(f"Began vectorizing {self}")
        with SamReader(xam_inp=self.bam_path,
                       end5=self.end5,
                       end3=self.end3,
                       isfullref=self.isfullref,
                       temp_dir=temp_dir,
                       n_procs=n_procs,
                       save_temp=save_temp) as reader:
            # Use integer division to round down the number of vectors
            # per batch to avoid exceeding the given batch size. But
            # also ensure that there is at least one vector per batch.
            vectors_per_batch = max(1, mib_to_bytes(batch_size) // self.length)
            # Compute for each batch the positions in the SAM file
            # (given by file.tell and used by file.seek) at which the
            # batch starts and stops.
            starts, stops = reader.list_batch_indexes(vectors_per_batch)
            n_batches = len(starts)
            batch_nums = list(range(n_batches))
            # Once the number of batches has been determined, a list of
            # new SAM file readers is created. Each is responsible for
            # converting one batch of SAM reads into one batch of
            # mutation vectors.
            readers = [SamReader(xam_inp=reader.sam_path,
                                 end5=self.end5,
                                 end3=self.end3,
                                 isfullref=self.isfullref,
                                 temp_dir=temp_dir,
                                 n_procs=n_procs,
                                 save_temp=save_temp)
                       for _ in batch_nums]
            # Create lists of the positional aguments for each reader.
            iter_args = list(zip(batch_nums, readers, starts, stops,
                                 strict=True))
            # Use the same keyword arguments (out_dir and other kwargs)
            # for every batch being vectorized.
            vectorize_batch = partial(self._vectorize_batch,
                                      out_dir=out_dir,
                                      **kwargs)
            if (pool_size := min(n_procs, n_batches)) > 1:
                # Process batches of records simultaneously in parallel.
                logger.debug(f"Initializing pool of {pool_size} processes")
                with Pool(pool_size) as pool:
                    logger.debug(f"Opened pool of {pool_size} processes")
                    results = list(pool.starmap(vectorize_batch, iter_args))
                logger.debug(f"Closed pool of {pool_size} processes")
            else:
                # Process batches of records one at a time in series.
                results = list(itsmap(vectorize_batch, iter_args))
        # The list of results contains, for each batch, a tuple of the
        # number of mutation vectors in the batch and the MD5 checksum
        # of the batch. Compute the total number of vectors and list all
        # the checksums.
        n_pass = 0
        n_fail = 0
        checksums: list[str] = list()
        for p, f, c in results:
            n_pass += p
            n_fail += f
            checksums.append(c)
        logger.info(f"Ended vectorizing {self}: {n_pass} pass, {n_fail} fail")
        return n_pass, n_fail, checksums

    def vectorize(self, /, *, rerun: bool, out_dir: Path, **kwargs):
        """ Compute a mutation vector for every record in a BAM file,
        write the vectors into one or more batch files, compute their
        checksums, and write a report summarizing the results. """
        report_path = self.get_report_path(out_dir)
        if self.outputs_valid(out_dir) and not rerun:
            logger.warning(f"Skipping vectorization of {self} because output "
                           f"files already exist. To rerun, add --rerun")
        else:
            # Compute the mutation vectors, write them to batch files,
            # and generate a report.
            try:
                # Vectorize the BAM file and time how long it takes.
                began = datetime.now()
                n_pass, n_fail, checksums = self._vectorize_bam(out_dir=out_dir,
                                                                **kwargs)
                ended = datetime.now()
                # Write a report of the vectorization.
                written = self._write_report(out_dir=out_dir,
                                             n_vectors=n_pass,
                                             n_readerr=n_fail,
                                             checksums=checksums,
                                             began=began,
                                             ended=ended)
                if written != report_path:
                    logger.critical(
                        "Intended and actual paths of report differ: "
                        f"{report_path} ≠ {written}")
            except Exception as error:
                # Alert that vectoring failed and return no report path.
                logger.critical(f"Failed to vectorize {self}: {error}")
                return None
        return report_path


class VectorReport(object):
    """
    Read and write a report about a mutational profile, including:
    - the sample, reference, and section
    - number of mutation vectors
    - number of mutation vector batch files and their checksums
    - beginning and ending time, duration, and speed of vectoring

    Examples
    --------
    >>> report = VectorReport(sample="dms2", ref="tmv-rna", end5=1, end3=20,
    ...                       seq=DNA(b"GTATTTTTACAACAATTACC"),
    ...                       n_vectors=10346, n_batches=2,
    ...                       checksums=["b47260fcad8be60470bee67862f187b4",
    ...                                  "098f40cfadc266ea5bc48ab2e18cdc95"],
    ...                       began=datetime.now(),
    ...                       ended=(sleep(1E-5), datetime.now())[-1])
    >>> str(report["seq"])
    'GTATTTTTACAACAATTACC'
    """

    class AbstractField(object):
        dtype: type
        key: str
        alias: str

        def __init__(self, value: Any):
            if not isinstance(value, self.dtype):
                raise TypeError(f"{self.__class__.__name__} expected a value "
                                f"of type '{self.dtype.__name__}' but got "
                                f"type '{type(value).__name__}'")
            self.value: Any = value

        @classmethod
        def parse(cls, value: str):
            # For the base class, just cast value to the expected type.
            return cls(cls.dtype(value))

        def __str__(self):
            return str(self.value)

    class AbstractStrField(AbstractField):
        """ String field """
        dtype = str

    class AbstractNumField(AbstractField):
        """ Numeric field with optional minimal and maximal values """
        dtype: int | float
        minimum: int | float | None = None
        maximum: int | float | None = None

        def __init__(self, value: int | float):
            super().__init__(value)
            # Verify bounds. Note that if value is NaN, the < and >
            # comparisons always return False, so the checks will pass,
            # which is the intended behavior. Do not check for
            # if not value >= self.minimum and
            # if not value <= self.maximum
            # because then the checks will fail for NaN values.
            if self.minimum is not None:
                if value < self.minimum:
                    raise ValueError(f"{self.alias} must be ≥ {self.minimum}, "
                                     f"but got {value}")
            if self.maximum is not None:
                if value > self.maximum:
                    raise ValueError(f"{self.alias} must be ≤ {self.maximum}, "
                                     f"but got {value}")

    class AbstractPosIntField(AbstractNumField):
        """ Positive integer """
        dtype, minimum = int, 1

    class AbstractNonNegIntField(AbstractNumField):
        """ Non-negative integer """
        dtype, minimum = int, 0

    class AbstractNonNegFloatField(AbstractNumField):
        """ Non-negative real number """
        dtype, minimum = float, 0.0

    class SampleField(AbstractStrField):
        key, alias = "sample", "Sample Name"

    class RefField(AbstractStrField):
        key, alias = "ref", "Reference Name"

    class End5Field(AbstractPosIntField):
        key, alias = "end5", "Section 5' End"

    class End3Field(AbstractPosIntField):
        key, alias = "end3", "Section 3' End"

    class LengthField(AbstractPosIntField):
        key, alias = "length", "Section Length"

    class AbstractDnaField(AbstractField):
        dtype = DNA

        @classmethod
        def parse(cls, value: str):
            # Need to encode the value to bytes before casting.
            return cls(cls.dtype(value.encode()))

        def __str__(self):
            # Need to decode the value from DNA to str.
            return self.value.decode()

    class SeqField(AbstractDnaField):
        key, alias = "seq", "Section Sequence"

    class BoolField(AbstractField):
        dtype = bool

        @classmethod
        def parse(cls, value: str):
            lower = value.strip().lower()
            if lower == "true":
                return cls(True)
            if lower == "false":
                return cls(False)
            raise ValueError(f"Cannot parse '{value}' as {cls.dtype.__name__}")

    class IsFullField(BoolField):
        key, alias = "isfullref", "Section is Full Reference"

    class NumVectorsField(AbstractNonNegIntField):
        key, alias = "n_vectors", "Reads Vectorized"

    class NumReadErrorsField(AbstractNonNegIntField):
        key, alias = "n_readerr", "Reads with Errors"

    class FracVectorizedField(AbstractNonNegFloatField):
        key, alias = "f_success", "Fraction Vectorized"

    class NumBatchesField(AbstractNonNegIntField):
        key, alias = "n_batches", "Batches"

    class ListStrField(AbstractField):
        dtype, delim = list, ", "

        @classmethod
        def parse(cls, value: str):
            return cls(value.strip().split(cls.delim) if value else list())

        def __str__(self):
            return self.delim.join(self.value)

    class ChecksumsField(ListStrField):
        key, alias = "checksums", "MD5 Checksums"

    class AbstractDateTimeField(AbstractField):
        dtype, dtfmt = datetime, "%Y-%m-%d %H:%M:%S.%f"

        @classmethod
        def parse(cls, value: str):
            return cls(cls.dtype.strptime(value, cls.dtfmt))

        def __str__(self):
            return self.value.strftime(self.dtfmt)

    class TimeBeganField(AbstractDateTimeField):
        key, alias = "began", "Began"

    class TimeEndedField(AbstractDateTimeField):
        key, alias = "ended", "Ended"

    class TimeTakenField(AbstractNonNegFloatField):
        key, alias = "taken", "Time taken (s)"

    class SpeedField(AbstractNonNegFloatField):
        key, alias = "speed", "Speed (1/s)"

    fields = [SampleField, RefField, End5Field, End3Field, LengthField,
              SeqField, IsFullField, NumVectorsField, NumReadErrorsField,
              FracVectorizedField, NumBatchesField, ChecksumsField,
              TimeBeganField, TimeEndedField, TimeTakenField, SpeedField]

    fields_by_key = {field.key: field for field in fields}

    fields_by_alias = {field.alias: field for field in fields}

    @staticmethod
    def compute_length(rep: VectorReport):
        return rep[rep.End3Field] - rep[rep.End5Field] + 1

    @staticmethod
    def compute_f_success(rep: VectorReport):
        try:
            return round((rep[rep.NumVectorsField]
                          / (rep[rep.NumVectorsField]
                             + rep[rep.NumReadErrorsField])
                          ), 5)
        except ZeroDivisionError:
            return float("nan")

    @staticmethod
    def compute_n_batches(rep: VectorReport):
        return len(rep[rep.ChecksumsField])

    @staticmethod
    def compute_time_taken(rep: VectorReport):
        dtime: timedelta = rep[rep.TimeEndedField] - rep[rep.TimeBeganField]
        return round(dtime.seconds + dtime.microseconds / 1000000, 3)

    @staticmethod
    def compute_speed(rep: VectorReport):
        try:
            return round(rep[rep.NumVectorsField] / rep[rep.TimeTakenField], 1)
        except ZeroDivisionError:
            return float("inf" if rep[rep.NumVectorsField] else "nan")

    compute_fields = {LengthField.key: compute_length,
                      FracVectorizedField.key: compute_f_success,
                      NumBatchesField.key: compute_n_batches,
                      TimeTakenField.key: compute_time_taken,
                      SpeedField.key: compute_speed}

    def __init__(self, /, **kwargs):
        # Fill fields using keyword arguments.
        missing = list()
        for key, field in self.fields_by_key.items():
            if (value := kwargs.get(key)) is None:
                if key not in self.compute_fields:
                    missing.append(key)
            else:
                self[field] = value
        # Raise an error if fields that cannot be computed are missing.
        if missing:
            raise TypeError("Missing the following required keyword arguments: "
                            + ", ".join(missing))
        # Compute any missing fields.
        for key, func in self.compute_fields.items():
            if key not in self:
                self[key] = func(self)

    def __getitem__(self, item: str | type[AbstractField]):
        if isclass(item) and issubclass(item, self.AbstractField):
            item = item.key
        return self.__getattribute__(item)

    def __setitem__(self, field: str | type[AbstractField], value: Any):
        if isinstance(field, str):
            field = self.fields_by_key[field]
        self.__setattr__(field.key, field(value).value)

    def __contains__(self, item: str | type[AbstractField]):
        if isclass(item) and issubclass(item, self.AbstractField):
            item = item.key
        return hasattr(self, item)

    def get_values_by_key(self, fields: Sequence[type[AbstractField]]):
        return {field.key: self[field] for field in fields}

    @cached_property
    def reader_dict(self):
        rdict = self.get_values_by_key([self.SampleField, self.RefField,
                                        self.End5Field, self.End3Field,
                                        self.IsFullField, self.NumVectorsField,
                                        self.SeqField, self.ChecksumsField])
        rdict["sect_seq"] = rdict.pop("seq")
        return rdict

    @cached_property
    def path_dict(self):
        return dict(module=path.Module.VECTOR,
                    **self.get_values_by_key([self.SampleField, self.RefField,
                                              self.End5Field, self.End3Field]))

    def find_invalid_batches(self, out_dir: str, validate_checksums: bool):
        """ Return all the batches of mutation vectors that either do not exist
        or do not match their expected checksums. """
        missing = list()
        badsum = list()
        for batch, checksum in enumerate(self[self.ChecksumsField]):
            file = path.MutVectorBatchFilePath(top=out_dir,
                                               **self.path_dict,
                                               batch=batch,
                                               ext=".orc")
            file_path = file.path
            if file_path.is_file():
                if validate_checksums and digest_file(file_path) != checksum:
                    # The batch file exists but does not match the checksum.
                    badsum.append(file)
            else:
                # The batch file does not exist.
                missing.append(file)
        return missing, badsum

    def save(self, out_dir: Path):
        dict_strs = {alias: str(field(self[field]))
                     for alias, field in self.fields_by_alias.items()}
        report_path = path.MutVectorReportFilePath(top=str(out_dir),
                                                   ext=".json",
                                                   **self.path_dict)
        report_path.path.parent.mkdir(parents=True, exist_ok=True)
        with open(report_path.path, "w") as f:
            json.dump(dict_strs, f)
        return report_path

    @classmethod
    def load(cls,
             report_file: str,
             validate_checksums: bool = True) -> VectorReport:
        """ Load a mutation vector report from a file. """
        with open(report_file, "r") as f:
            dict_strs = json.load(f)
        dict_vals = dict()
        for alias, value in dict_strs.items():
            field = cls.fields_by_alias[alias]
            dict_vals[field.key] = field.parse(value).value
        report = cls(**dict_vals)
        report_path = path.MutVectorReportFilePath.parse(report_file)
        if (report_path.sample != report[cls.SampleField]
                or report_path.ref != report[cls.RefField]
                or report_path.end5 != report[cls.End5Field]
                or report_path.end3 != report[cls.End3Field]):
            raise ValueError(f"Report fields do not match path: {report_path}")
        missing, badsum = report.find_invalid_batches(report_path.top,
                                                      validate_checksums)
        if missing:
            raise FileNotFoundError(f"Batches: {', '.join(map(str, missing))}")
        if badsum:
            raise ValueError(f"Bad MD5 sums: {', '.join(map(str, badsum))}")
        return report


class VectorReader(MutationalProfile):
    INDEX_COL = "__index_level_0__"

    def __init__(self, /, *,
                 out_dir: Path,
                 n_vectors: int,
                 checksums: list[str],
                 **kwargs):
        super().__init__(**kwargs)
        self.out_dir = out_dir
        self.n_vectors = n_vectors
        self.checksums = checksums

    @property
    def n_batches(self):
        return len(self.checksums)

    @property
    def batch_nums(self) -> list[int]:
        """ Return a list of all batch numbers. """
        return list(range(self.n_batches))

    def get_mv_batch_paths(self, out_dir: Path):
        """ Return the path of every mutation vector batch file. """
        return [self.get_mv_batch_path(out_dir, batch)
                for batch in self.batch_nums]

    @property
    def shape(self):
        return self.n_vectors, self.length

    def get_batch(self,
                  batch: int,
                  positions: Sequence[int] | None = None,
                  numeric: bool = False):
        """
        Return the mutation vectors from one batch. Optionally, select
        a subset of the columns of the mutation vectors.

        Parameters
        ----------
        batch: int (≥ 0)
            Number of the batch of mutation vectors
        positions: sequence[int] | None (default: None)
            If given, use only these positions from the mutation vectors;
            otherwise, use all positions.
        numeric: bool (default: False)
            Whether to convert the columns from {base}{position} strings
            to numeric (integer) values of just the positions, e.g.
            ['A14', 'G15', ...] if False, [14, 15, ...] if True

        Return
        ------
        DataFrame
            Mutation vectors; each row is a vector indexed by its name,
            each column a position indexed by its positional number
        """
        if batch not in self.batch_nums:
            raise ValueError(f"Invalid batch number: {batch} "
                             f"(expected one of {list(self.batch_nums)})")
        # Determine which columns to read from the file.
        if positions is None:
            # Read all columns in the file.
            columns = None
        else:
            # Read only columns corresponding to the given positions.
            subseq = self.subseq(positions)
            columns = [self.INDEX_COL] + self.seq_pos_to_cols(subseq, positions)
        # Read the vectors from the ORC file using PyArrow as backend.
        vectors = pd.read_orc(self.get_mv_batch_path(self.out_dir, batch).path,
                              columns=columns)
        # Remove the column of read names and set it as the index.
        vectors.set_index(self.INDEX_COL, drop=True, inplace=True)
        # Convert the index from bytes to str and give it a name.
        vectors.set_index(pd.Index(vectors.index.map(bytes.decode),
                                   name="Read Name"),
                          inplace=True)
        if numeric:
            # Convert the remaining columns to their integer positions.
            vectors.columns = positions if positions else self.positions
        # The vectors are stored as signed 8-bit integers (np.int8) and
        # must be cast to unsigned 8-bit integers (np.uint8) so that the
        # bitwise operations work. This step must be doneafter removing
        # the column of read names (which cannot be cast to np.uint8).
        return vectors.astype(np.uint8, copy=False)

    def get_all_batches(self,
                        positions: Sequence[int] | None = None,
                        numeric: bool = True):
        """
        Yield every batch of mutation vectors.

        Parameters
        ----------
        positions: sequence[int] | None (default: None)
            If given, use only these position from the mutation vectors;
            otherwise, use all positions.
        numeric: bool (default: False)
            Whether to convert the columns from {base}{position} strings
            to numeric (integer) values of just the positions, e.g.
            ['A14', 'G15', ...] if False, [14, 15, ...] if True

        Yield
        -----
        DataFrame
            Mutation vectors; each row is a vector indexed by its name,
            each column a position indexed by its base and number
        """
        for batch in self.batch_nums:
            yield self.get_batch(batch, positions, numeric)

    def get_all_vectors(self,
                        positions: Sequence[int] | None = None,
                        numeric: bool = True):
        """
        Return all mutation vectors for this vector reader. Note that
        reading all vectors could take more than the available memory
        and cause the program to crash. Thus, use this method only if
        all vectors will fit into memory. Otherwise, use the method
        ```get_all_batches``` to process the vectors in small batches.

        Parameters
        ----------
        positions: sequence[int] | None (default: None)
            If given, use only these position from the mutation vectors;
            otherwise, use all positions.
        numeric: bool (default: False)
            Whether to convert the columns from {base}{position} strings
            to numeric (integer) values of just the positions, e.g.
            ['A14', 'G15', ...] if False, [14, 15, ...] if True

        Return
        ------
        DataFrame
            Mutation vectors; each row is a vector indexed by its name,
            each column a position indexed by its base and number
        """
        # If there are no vectors, then return an empty DataFrame.
        if self.n_batches == 0:
            return pd.DataFrame(columns=(self.positions if positions is None
                                         else positions), dtype=int)
        # Load and concatenate every vector batch into one DataFrame.
        return pd.concat(self.get_all_batches(positions, numeric), axis=0)

    @classmethod
    def query_vectors(cls, /,
                      vectors: pd.DataFrame,
                      query: int, *,
                      subsets: bool = False,
                      supersets: bool = False) -> pd.DataFrame:
        """
        Return a boolean array of the same shape as vectors where
        element i,j is True if and only if the byte at element i,j of
        vectors matches the given query byte. By default, a byte in
        vectors matches if it equals the query byte and is not blank
        (i.e. 00000000). Matches can be extended to include bitwise
        subsets and supersets of query by setting the corresponding
        parameters to True. Blank bytes in vectors never match.

        Parameters
        ----------
        vectors: DataFrame
            Mutation vectors
        query: int
            Byte to query; must be in range [0, 255]
        subsets: bool
            Whether to count non-blank bitwise subsets of the query
        supersets: bool
            Whether to count non-blank bitwise supersets of the query

        Return
        ------
        DataFrame
            Boolean type DataFrame of the same shape as vectors where
            each element is True if the element at the same position in
            vectors matched the query and False otherwise
        """
        if not isinstance(query, int):
            raise TypeError(
                f"Expected query of type int, but got {type(query).__name__}")
        if not BLANK <= query <= EVERY:
            raise ValueError(
                f"Expected query in range {BLANK} - {EVERY}, but got {query}")
        if supersets and subsets:
            # Count both supersets and subsets.
            return (cls.query_vectors(vectors, query, supersets=True)
                    | cls.query_vectors(vectors, query, subsets=True))
        if supersets:
            # Non-blank vector bytes that are matches and supersets of
            # the query byte count.
            if query == BLANK:
                # If query is BLANK (00000000), then every non-blank
                # byte is a superset.
                return vectors.astype(bool, copy=True)
            if query == EVERY:
                # If the query is EVERY (11111111), then no bitwise
                # supersets exist. Since subsets do not count, only
                # exact matches count.
                return cls.query_vectors(vectors, query)
            # No shortcut method can be used, so the supersets must be
            # computed explicitly. A vector byte is a match or superset
            # of the query byte iff both of the following are true:
            # - The bitwise intersection of the vector and query bytes
            #   equals the query byte, meaning that every bit set to 1
            #   in the query byte is also set to 1 in the vector byte,
            #   and thus the vector byte is a superset of the query.
            #   Equivalently, the union equals the vector byte.
            # - The vector byte is not blank. But since the query byte
            #   is not blank, if a vector byte satisfies the first
            #   condition, then it must also satisfy this one, so this
            #   latter condition need not be checked.
            return np.equal(np.bitwise_and(vectors, query), query)
        if query == BLANK:
            # If supersets do not count, then only matches and subsets
            # count. But if the query is BLANK (00000000), then there
            # are no subsets, and matches do not count because blank
            # vector bytes never count. Thus, no byte counts.
            return pd.DataFrame(False, dtype=bool,
                                index=vectors.index,
                                columns=vectors.columns)
        if subsets:
            # Non-blank vector bytes that are matches and subsets of the
            # query byte count.
            if query == EVERY:
                # If query is EVERY (11111111), then every non-blank
                # byte is a subset.
                return vectors.astype(bool, copy=True)
            if (query & (query - 1)) == 0:
                # If query is a power of 2, then it has exactly one bit
                # set to 1. Thus, the only possible subset of the query
                # is the blank byte, which never counts. Since supersets
                # do not count either, only exact matches count.
                return cls.query_vectors(vectors, query)
            # No shortcut method can be used, so the subsets must be
            # computed explicitly. A vector byte is a match or subset of
            # the query byte iff both of the following are true:
            # - The bitwise union of the vector and query bytes equals
            #   the query byte, meaning that there are no bits set to 1
            #   in the vector byte that are not 1 in the query byte,
            #   and thus the vector byte is a subset of the query.
            #   Equivalently, the intersection equals the vector byte.
            # - The vector byte is not blank.
            return (vectors.astype(bool, copy=False)
                    & np.equal(np.bitwise_or(vectors, query), query))
        # If neither subsets nor supersets count and query is not BLANK,
        # then count vector bytes that match the query byte exactly.
        return np.equal(vectors, query)

    @cache
    def count_muts_by_vec(self, /,
                          query: int, *,
                          subsets: bool = False,
                          supersets: bool = False,
                          positions: Sequence[int] | None = None) -> pd.Series:
        """
        Return the number of mutations that match the query for each
        vector in the mutational profile.

        Parameters
        ----------
        query: int
            Byte to query; must be in range [0, 255]
        subsets: bool
            Whether to count non-blank bitwise subsets of the query
        supersets: bool
            Whether to count non-blank bitwise supersets of the query
        positions: sequence[int] | None
            Use only these positions from the mutation vectors; if None,
            then use all positions.

        Return
        ------
        Series
            Number of mutations matching the query in each vector
        """
        # If there are no vectors, then return an empty Series.
        if self.n_batches == 0:
            return pd.Series([], dtype=int)
        # Initialize empty list to count the mutations in each vector.
        counts = list()
        # Iterate over all batches of vectors.
        for vectors in self.get_all_batches(positions):
            # Count the number of mutations in each vector in the batch
            # and append them to the list of counts.
            counts.append(self.query_vectors(vectors,
                                             query,
                                             subsets=subsets,
                                             supersets=supersets).sum(axis=1))
        # Concatenate and return the number of mutations in each vector
        # among all batches.
        return pd.concat(counts, axis=0)

    @cache
    def count_muts_by_pos(self, /,
                          query: int, *,
                          subsets: bool = False,
                          supersets: bool = False,
                          positions: Sequence[int] | None = None,
                          numeric: bool = True) -> pd.Series:
        """
        Return the number of mutations that match the query at each
        position in the mutational profile.

        Parameters
        ----------
        query: int
            Byte to query; must be in range [0, 255]
        subsets: bool
            Whether to count non-blank bitwise subsets of the query
        supersets: bool
            Whether to count non-blank bitwise supersets of the query
        positions: sequence[int] | None
            Use only these positions from the mutation vectors; if None,
            then use all positions.
        numeric: bool (default: False)
            Whether to convert the columns from {base}{position} strings
            to numeric (integer) values of just the positions, e.g.
            ['A14', 'G15', ...] if False, [14, 15, ...] if True

        Return
        ------
        Series
            Number of mutations matching the query at each position
        """
        # Initialize empty Series to count mutations at each position.
        counts = pd.Series(np.zeros(self.length, dtype=int),
                           index=(self.positions if positions is None
                                  else positions))
        # Iterate over all batches of vectors.
        for vectors in self.get_all_batches(positions, numeric):
            # Add the number of mutations at each position in the batch
            # to the cumulative count of mutations at each position.
            counts += self.query_vectors(vectors,
                                         query,
                                         subsets=subsets,
                                         supersets=supersets).sum(axis=0)
        return counts

    def get_cluster_mus(self, /,
                        membership: pd.DataFrame,
                        query: int, *,
                        subsets: bool = False,
                        supersets: bool = False,
                        positions: Sequence[int] | None = None,
                        numeric: bool = False) -> pd.DataFrame:
        """
        Calculate the Mutation fraction at each position in a mutational
        profile for one or more clusters.

        Parameters
        ----------
        membership: DataFrame
            Cluster membership: each index (i) is the name of a read,
            each column (k) the name of a cluster, and each value (i, k)
            the likelihood that read (i) came from cluster (k).
        query: int
            Byte to query; must be in range [0, 255]
        subsets: bool
            Whether to count non-blank bitwise subsets of the query
        supersets: bool
            Whether to count non-blank bitwise supersets of the query
        positions: sequence[int] | None
            Use only these positions from the mutation vectors; if None,
            then use all positions.
        numeric: bool (default: False)
            Whether to convert the columns from {base}{position} strings
            to numeric (integer) values of just the positions, e.g.
            ['A14', 'G15', ...] if False, [14, 15, ...] if True

        Return
        ------
        DataFrame
            Mutation rates: each index (j) is a position in the profile,
            each column (k) the name of a cluster, and each value (j, k)
            the Mutation fraction at posision (j) in cluster (k).

        Explanation
        -----------
        To limit memory usage, this function is implemented on a single
        line. This is what each step does:

        1.  Read all mutation vectors into a DataFrame. Assume that if
            the reads were clustered, then they can all fit into memory
            at once. It is much easier to compute products of indexed
            DataFrames all at once than by summing over chunks:

            mutvectors = self.get_all_vectors(positions, numeric)

        2.  Compute the bit vectors: a matrix of boolean values that
            indicate whether each read (row) is mutated at each position
            (column). The matrix has the same shape as mutvectors. The
            query determines which mutations count as True or False:

            bitvectors = self._query_vectors(mutvectors, query,
                                             subsets=subsets,
                                             supersets=supersets)

        3.  Compute the weighted number of mutations at each position in
            each cluster, weighed by the likelihood that each read came
            from each cluster. The resulting matrix has a row for each
            position and a column for each cluster:

            mutsums = bitvectors.T.dot(membership)

        4.  Compute the weighted number of reads in each cluster by just
            summing the likelihood that each read came from the cluster:

            readsums = membership.sum(axis=0)

        5.  Compute the mutation rates for each cluster by dividing the
            cluster-weighted number of mutations at each position by the
            weighted number of reads in the cluster:

            return mutsums / readsums
        """
        return (self.query_vectors(self.get_all_vectors(positions, numeric),
                                   query,
                                   subsets=subsets,
                                   supersets=supersets).T.dot(membership)
                / membership.sum(axis=0))

    @classmethod
    def load(cls, report_file: str, validate_checksums: bool = True):
        if not (report := VectorReport.load(report_file, validate_checksums)):
            logger.critical(f"Failed to load report from {report_file}")
            return
        return cls(out_dir=Path(
            path.MutVectorReportFilePath.parse(report_file).top),
            **report.reader_dict)


def get_min_qual(min_phred: int, phred_enc: int):
    """
    Return the minimum quality for a base in a read to be considered
    informative, as the ASCII integer corresponding to the character
    in the FASTQ file that is the minimum valid quality.

    Return
    ------
    int
        The ASCII value corresponding to the character in the FASTQ
        file read quality string that represents the minimum quality
        for a base to be considered informative

    Examples
    --------
    For example, if the minimum Phred score (```min_phred```) that
    is accepted as informative is 20, and the Phred encoding of the
    FASTQ file (```phred_enc```) is 33 (i.e. ASCII+33), then the
    minimum quality as an ASCII integer (```min_qual```) is 20 + 33
    = 53, which is character '5'. If ```min_phred``` were 37, then
    ```min_qual``` would be 37 + 33 = 70, which is character 'F'.
    """
    return min_phred + phred_enc


def get_sections(ref_seqs: dict[str, DNA], *,
                 coords: tuple[tuple[str, int, int]],
                 primers: tuple[tuple[str, DNA, DNA]],
                 primer_gap: int,
                 autosect: bool):
    """ Return all the sections corresponding to the given coordinates
    and/or primers in the given reference sequences. """
    sections: dict[str, list[SectionFinder]] = defaultdict(list)

    def add_section(section: SectionFinder):
        if any(section == other for other in sections[section.ref]):
            logger.warning(f"Skipping duplicate section: {section.ref_coords}")
        sections[section.ref].append(section)

    for ref, first, last in coords:
        add_section(SectionFinder(ref_seq=ref_seqs.get(ref), ref=ref,
                                  end5=first, end3=last, primer_gap=primer_gap))
    for ref, fwd, rev in primers:
        add_section(SectionFinder(ref_seq=ref_seqs.get(ref), ref=ref,
                                  fwd=fwd, rev=rev, primer_gap=primer_gap))
    if autosect:
        for ref, seq in ref_seqs.items():
            if ref not in sections:
                add_section(SectionFinder(ref_seq=seq, ref=ref,
                                          primer_gap=primer_gap))
    if missing := set(ref_seqs) - set(sections):
        logger.warning(
            f"No sections given for references: {', '.join(missing)}")
    return sections


def _build_temp_bam_index(temp_dir: Path,
                          bam_inp: path.OneRefAlignmentInFilePath,
                          n_procs: int):
    """ Create an index a temporary BAM file. """
    bam_file = path.OneRefAlignmentTempFilePath(top=str(temp_dir),
                                                module=path.Module.VECTOR,
                                                step=path.Step.VECTOR_BAMS,
                                                sample=bam_inp.sample,
                                                ref=bam_inp.ref,
                                                ext=path.BAM_EXT).path
    # Create the new directory.
    logger.debug(f"Creating directory {bam_file.parent}")
    bam_file.parent.mkdir(parents=True, exist_ok=True)
    try:
        # Link the new BAM file to the input BAM path.
        bam_file.symlink_to(bam_inp.path)
        logger.debug(f"Linked {bam_file} to point at {bam_inp.path}")
    except FileExistsError:
        # If the path already exists, check if it is a symbolic link
        # that points to the correct BAM file.
        if bam_file.is_symlink() and bam_file.readlink() == bam_inp.path:
            # If so, then it will function properly, but it should not
            # have existed, so issue a warning.
            logger.warning(f"Link already exists: {bam_file}  -->  {bam_inp}")
        else:
            # Otherwise, it will not function properly. Re-raise the
            # error rather than deleting this arbitrary file.
            raise
    # Build the temporary index.
    index_bam(bam_file, n_procs)
    return bam_file


def get_writers(fasta: Path,
                bam_files: list[Path], *,
                temp_dir: Path,
                n_procs: int,
                **kwargs):
    logger.info("Began creating vector writers")
    ref_seqs = dict(parse_fasta(fasta))
    sections = get_sections(ref_seqs, **kwargs)
    writers: dict[tuple, VectorWriter] = dict()
    temp_files = list()
    for bam_file in bam_files:
        logger.debug(f"Creating vector writers for {bam_file}")
        try:
            # Parse the fields of the input BAM file.
            bam_inp = path.OneRefAlignmentInFilePath.parse(bam_file)
            # Check whether the input BAM file has an index.
            if not bam_file.with_suffix(path.BAI_EXT).is_file():
                # An index must be built for the BAM file. To avoid altering
                # the input directory, build the index in a new directory.
                bam_file = _build_temp_bam_index(temp_dir, bam_inp, n_procs)
                temp_files.append(bam_file)
                temp_files.append(bam_file.with_suffix(path.BAI_EXT))
            for section in sections[bam_inp.ref]:
                if section.ref != bam_inp.ref:
                    # This should never happen. Handling just in case.
                    raise ValueError(f"Section {section} of {bam_inp} had a "
                                     f"reference '{section.ref}' that did not "
                                     f"match the BAM file '{bam_inp.ref}'.")
                writer = VectorWriter(bam_file=bam_file,
                                      ref_seq=ref_seqs[bam_inp.ref],
                                      end5=section.end5,
                                      end3=section.end3)
                if writer.tag in writers:
                    logger.warning("Skipping duplicate mutational profile: "
                                   f"{writer}.")
                    continue
                writers[writer.tag] = writer
                logger.debug(f"Created vector writer for {bam_file} {section}")
        except Exception as error:
            logger.critical(f"Error in BAM file {bam_file}: {error}")
    logger.info(f"Ended creating {len(writers)} vector writers")
    return list(writers.values()), temp_files


def generate_profiles(writers: list[VectorWriter], *,
                      phred_enc: int,
                      min_phred: int,
                      max_procs: int,
                      parallel: bool,
                      **kwargs) -> tuple[str, ...]:
    """ Generate mutational profiles of one or more vector writers. """
    logger.info("Began generating mutational profiles")
    n_profiles = len(writers)
    if n_profiles == 0:
        raise ValueError("No BAM files and/or sections specified")
    # Determine method of parallelization. Do not use hybrid mode, which
    # would try to process multiple SAM files in parallel and use more
    # than one processe for each file. Python ```multiprocessing.Pool```
    # forbids a daemon process (one for each SAM file in parallel) from
    # spawning additional processes.
    n_tasks_parallel, n_procs_per_task = get_num_parallel(n_profiles,
                                                          max_procs,
                                                          parallel)
    # List the arguments of VectorWriter.vectorize for each writer.
    iter_args = [(writer,) for writer in writers]
    vectorize = partial(VectorWriter.vectorize,
                        phred_enc=phred_enc,
                        min_phred=min_phred,
                        n_procs=n_procs_per_task,
                        **kwargs)
    # Call the vectorize method of each writer, passing args.
    if n_tasks_parallel > 1:
        logger.debug(f"Initializing pool of {n_tasks_parallel} processes")
        with Pool(n_tasks_parallel) as pool:
            logger.debug(f"Opened pool of {n_tasks_parallel} processes")
            report_files = tuple(pool.starmap(vectorize, iter_args))
        logger.debug(f"Closed pool of {n_tasks_parallel} processes")
    else:
        report_files = tuple(itsmap(vectorize, iter_args))
    # Filter out any None values (indicating failure), convert report
    # paths to a tuple of strings, and return.
    reports = tuple(map(str, filter(None, report_files)))
    logger.info(f"Ended generating mutational profiles: {len(reports)} pass, "
                f"{len(report_files) - len(reports)} fail")
    return reports


vector_trans_table = bytes.maketrans(*map(b"".join, zip(*[(
    i.to_bytes(length=1, byteorder=byteorder),
    (
        b"." if i == BLANK
        else b"~" if i == MATCH
        else b"/" if i == DELET
        else b"{" if i == (INS_5 | MATCH)
        else b"}" if i == (INS_3 | MATCH)
        else b"A" if i == SUB_A
        else b"C" if i == SUB_C
        else b"G" if i == SUB_G
        else b"T" if i == SUB_T
        else b"?"
    )
) for i in range(256)])))


def trans_vectors_iter(vectors: pd.DataFrame):
    for index, row in zip(vectors.index, vectors.values, strict=True):
        vector = row.tobytes(order='C').translate(vector_trans_table).decode()
        yield f"{index}\t{vector}\n"


def trans_vectors_block(vectors: pd.DataFrame, reference: bool = False):
    lines = trans_vectors_iter(vectors)
    if reference:
        # Display the reference sequence above the vectors.
        try:
            # Get the reference sequence from the column names.
            seq, _ = VectorReader.cols_to_seq_pos(vectors.columns.tolist())
            # Prepend the reference sequence to the lines of vectors.
            lines = chain([f"Reference\t{seq.decode()}\n"], lines)
        except Exception as error:
            logger.error(f"Could not determine sequence from columns of the "
                         f"vectors (perhaps you used numeric=True): {error} ")
    return "".join(lines)
